# Copyright (C) 2013 by Kevin L. Mitchell <klmitch@mit.edu>
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

import operator

import mock

from policies import instructions

import tests


class InstructionForTest(instructions.AbstractInstruction):
    def __call__(self, ctxt):
        pass

    def __repr__(self):
        return ''

    def __hash__(self, *elems):
        return super(InstructionForTest, self).__hash__(*elems)

    def __eq__(self, other):
        return super(InstructionForTest, self).__eq__(other)


class Instruction2ForTest(InstructionForTest):
    pass


class TestAbstractInstruction(tests.TestCase):
    def test_eq(self):
        one = InstructionForTest()
        two = InstructionForTest()
        three = Instruction2ForTest()

        self.assertTrue(one.__eq__(two))
        self.assertFalse(one.__eq__(three))

    def test_ne(self):
        one = InstructionForTest()
        two = InstructionForTest()
        three = Instruction2ForTest()

        self.assertFalse(one.__ne__(two))
        self.assertTrue(one.__ne__(three))

    def test_hash(self):
        one = InstructionForTest()

        self.assertEqual(hash(one), hash((InstructionForTest,)))
        self.assertEqual(one.__hash__(1, 2, 3),
                         hash((InstructionForTest, 1, 2, 3)))


class TestInstructions(tests.TestCase):
    @mock.patch.object(instructions.Instructions, '_linearize',
                       return_value=[1, 2, 3])
    def test_init(self, mock_linearize):
        insts = instructions.Instructions([3, 2, 1])

        self.assertEqual(insts.instructions, (1, 2, 3))
        mock_linearize.assert_called_once_with([3, 2, 1])

    @mock.patch.object(instructions.Instructions, '_linearize',
                       side_effect=lambda x: x)
    def test_repr(self, mock_linearize):
        insts = instructions.Instructions([1, 2, 3])

        self.assertEqual(repr(insts), 'Instructions((1, 2, 3))')

    @mock.patch.object(instructions.Instructions, '_linearize',
                       side_effect=lambda x: x)
    def test_call(self, mock_linearize):
        calls_obj = mock.Mock()
        insts = instructions.Instructions([calls_obj.one, calls_obj.two,
                                           calls_obj.three, calls_obj.four])

        insts('ctxt')

        calls_obj.assert_has_calls([
            mock.call.one('ctxt'),
            mock.call.two('ctxt'),
            mock.call.three('ctxt'),
            mock.call.four('ctxt'),
        ])

    @mock.patch.object(instructions.Instructions, '_linearize',
                       side_effect=lambda x: x)
    def test_call_no_authz(self, mock_linearize):
        calls_obj = mock.Mock()
        insts = instructions.Instructions([calls_obj.one, calls_obj.two,
                                           instructions.set_authz,
                                           calls_obj.three, calls_obj.four])

        insts('ctxt', True)

        calls_obj.assert_has_calls([
            mock.call.one('ctxt'),
            mock.call.two('ctxt'),
        ])
        self.assertEqual(len(calls_obj.method_calls), 2)

    @mock.patch.object(instructions.Instructions, '_linearize',
                       side_effect=lambda x: x)
    def test_hash(self, mock_linearize):
        insts = instructions.Instructions([1, 2, 3])

        self.assertEqual(hash(insts),
                         hash((instructions.Instructions, 1, 2, 3)))

    @mock.patch.object(instructions.Instructions, '_linearize',
                       side_effect=lambda x: x)
    def test_eq(self, mock_linearize):
        class Instructions2(instructions.Instructions):
            pass

        insts1 = instructions.Instructions([1, 2, 3])
        insts2 = instructions.Instructions([1, 2, 3])
        insts3 = instructions.Instructions([3, 2, 1])
        insts4 = Instructions2([1, 2, 3])

        self.assertTrue(insts1.__eq__(insts2))
        self.assertFalse(insts1.__eq__(insts3))
        self.assertFalse(insts1.__eq__(insts4))

    def test_linearize(self):
        with mock.patch.object(instructions.Instructions, '_linearize',
                               side_effect=lambda x: x):
            insts1 = instructions.Instructions([1, 2, 3])
            insts2 = instructions.Instructions(['a', 'b', 'c'])
            insts3 = instructions.Instructions([insts2, 'd', 'e'])
        feed = [9, 8, insts1, 7, 6, insts3, 5, 4]

        result = list(instructions.Instructions._linearize(feed))

        self.assertEqual(result, [
            9, 8, 1, 2, 3, 7, 6, 'a', 'b', 'c', 'd', 'e', 5, 4
        ])


class TestConstant(tests.TestCase):
    def test_init(self):
        constant = instructions.Constant('value')

        self.assertEqual(constant.value, 'value')

    def test_repr(self):
        constant = instructions.Constant('value')

        self.assertEqual(repr(constant), "Constant('value')")

    def test_call(self):
        ctxt = mock.Mock(stack=[])
        constant = instructions.Constant('value')

        constant(ctxt)

        self.assertEqual(ctxt.stack, ['value'])

    def test_hash(self):
        constant = instructions.Constant('value')

        self.assertEqual(hash(constant),
                         hash((instructions.Constant, 'value')))

    def test_eq(self):
        class Constant2(instructions.Constant):
            pass

        constant1 = instructions.Constant('value')
        constant2 = instructions.Constant('value')
        constant3 = instructions.Constant('other')
        constant4 = Constant2('value')

        self.assertTrue(constant1.__eq__(constant2))
        self.assertFalse(constant1.__eq__(constant3))
        self.assertFalse(constant1.__eq__(constant4))


class TestAttribute(tests.TestCase):
    def test_init(self):
        attribute = instructions.Attribute('attr')

        self.assertEqual(attribute.attribute, 'attr')

    def test_repr(self):
        attribute = instructions.Attribute('attr')

        self.assertEqual(repr(attribute), "Attribute('attr')")

    def test_call(self):
        ctxt = mock.Mock(stack=[mock.Mock(attr='value')])
        attribute = instructions.Attribute('attr')

        attribute(ctxt)

        self.assertEqual(ctxt.stack, ['value'])

    def test_hash(self):
        attribute = instructions.Attribute('attr')

        self.assertEqual(hash(attribute),
                         hash((instructions.Attribute, 'attr')))

    def test_eq(self):
        class Attribute2(instructions.Attribute):
            pass

        attribute1 = instructions.Attribute('attr')
        attribute2 = instructions.Attribute('attr')
        attribute3 = instructions.Attribute('other')
        attribute4 = Attribute2('attr')

        self.assertTrue(attribute1.__eq__(attribute2))
        self.assertFalse(attribute1.__eq__(attribute3))
        self.assertFalse(attribute1.__eq__(attribute4))


class TestIdent(tests.TestCase):
    def test_init(self):
        ident = instructions.Ident('ident')

        self.assertEqual(ident.ident, 'ident')

    def test_repr(self):
        ident = instructions.Ident('ident')

        self.assertEqual(repr(ident), "Ident('ident')")

    def test_call(self):
        ctxt = mock.Mock(**{
            'stack': [],
            'resolve.return_value': 'value',
        })
        ident = instructions.Ident('ident')

        ident(ctxt)

        self.assertEqual(ctxt.stack, ['value'])

    def test_hash(self):
        ident = instructions.Ident('ident')

        self.assertEqual(hash(ident),
                         hash((instructions.Ident, 'ident')))

    def test_eq(self):
        class Ident2(instructions.Ident):
            pass

        ident1 = instructions.Ident('ident')
        ident2 = instructions.Ident('ident')
        ident3 = instructions.Ident('other')
        ident4 = Ident2('ident')

        self.assertTrue(ident1.__eq__(ident2))
        self.assertFalse(ident1.__eq__(ident3))
        self.assertFalse(ident1.__eq__(ident4))


class OperatorForTest(instructions.Operator):
    def op(self, *args):
        return args


class Operator2ForTest(OperatorForTest):
    pass


class TestOperator(tests.TestCase):
    def test_init(self):
        op = OperatorForTest(5, 'opstr')

        self.assertEqual(op.count, 5)
        self.assertEqual(op.opstr, 'opstr')

    def test_repr(self):
        op = OperatorForTest(5, 'opstr')

        self.assertEqual(repr(op), "OperatorForTest(5, 'opstr')")

    def test_call(self):
        ctxt = mock.Mock(stack=[1, 2, 3, 4, 5])
        op = OperatorForTest(3, 'opstr')

        op(ctxt)

        self.assertEqual(ctxt.stack, [1, 2, (3, 4, 5)])

    def test_hash(self):
        op_func = lambda x: x
        op = OperatorForTest(5, 'opstr')

        self.assertEqual(hash(op),
                         hash((OperatorForTest, 5)))
        self.assertEqual(op.__hash__(op_func),
                         hash((OperatorForTest, 5, op_func)))

    def test_eq(self):
        op1 = OperatorForTest(5, 'op1')
        op2 = OperatorForTest(5, 'op2')
        op3 = OperatorForTest(3, 'op3')
        op4 = Operator2ForTest(5, 'op4')

        self.assertTrue(op1.__eq__(op2))
        self.assertFalse(op1.__eq__(op3))
        self.assertFalse(op1.__eq__(op4))

    def test_fold_constant(self):
        elems = [instructions.Constant(i) for i in range(3)]
        op = OperatorForTest(3, 'opstr')

        result = op.fold(elems)

        self.assertEqual(len(result), 1)
        self.assertTrue(isinstance(result[0], instructions.Constant))
        self.assertEqual(result[0].value, tuple(i.value for i in elems))

    def test_fold_nonconstant(self):
        elems = [instructions.Constant(i) for i in range(2)]
        elems += [instructions.Ident('spam')]
        op = OperatorForTest(3, 'opstr')

        result = op.fold(elems)

        self.assertEqual(len(result), 1)
        self.assertTrue(isinstance(result[0], instructions.Instructions))
        self.assertEqual(result[0].instructions, tuple(elems + [op]))


class TestGenericOperator(tests.TestCase):
    def test_init(self):
        gen_op = instructions.GenericOperator(3, 'op', 'opstr')

        self.assertEqual(gen_op.count, 3)
        self.assertEqual(gen_op.opstr, 'opstr')
        self.assertEqual(gen_op._op, 'op')

    def test_hash(self):
        gen_op = instructions.GenericOperator(3, 'op', 'opstr')

        self.assertEqual(hash(gen_op),
                         hash((instructions.GenericOperator, 3, 'op')))

    def test_eq(self):
        class GenericOperator2(instructions.GenericOperator):
            pass

        gen_op1 = instructions.GenericOperator(3, 'op', 'op1')
        gen_op2 = instructions.GenericOperator(3, 'op', 'op2')
        gen_op3 = instructions.GenericOperator(2, 'op', 'op3')
        gen_op4 = instructions.GenericOperator(3, 'other', 'op4')
        gen_op5 = GenericOperator2(3, 'op', 'op5')

        self.assertTrue(gen_op1.__eq__(gen_op2))
        self.assertFalse(gen_op1.__eq__(gen_op3))
        self.assertFalse(gen_op1.__eq__(gen_op4))
        self.assertFalse(gen_op1.__eq__(gen_op5))

    def test_op(self):
        op = mock.Mock(return_value='value')
        gen_op = instructions.GenericOperator(3, op, 'opstr')

        result = gen_op.op(1, 2, 3)

        self.assertEqual(result, 'value')
        op.assert_called_once_with(1, 2, 3)


class TestTrinaryOperator(tests.TestCase):
    def test_init(self):
        trinary = instructions.TrinaryOperator()

        self.assertEqual(trinary.count, 3)
        self.assertEqual(trinary.opstr, 'if/else')

    def test_op(self):
        trinary = instructions.TrinaryOperator()

        self.assertEqual(trinary.op(1, 'true', 'false'), 'true')
        self.assertEqual(trinary.op(0, 'true', 'false'), 'false')

    def test_fold_constant_true(self):
        elems = [instructions.Constant(True), instructions.Ident('a'),
                 instructions.Ident('b')]
        trinary = instructions.TrinaryOperator()

        result = trinary.fold(elems)

        self.assertEqual(len(result), 1)
        self.assertEqual(result, [elems[1]])

    def test_fold_constant_false(self):
        elems = [instructions.Constant(False), instructions.Ident('a'),
                 instructions.Ident('b')]
        trinary = instructions.TrinaryOperator()

        result = trinary.fold(elems)

        self.assertEqual(len(result), 1)
        self.assertEqual(result, [elems[2]])

    def test_fold_nonconstant(self):
        elems = [instructions.Ident('a'), instructions.Constant('true'),
                 instructions.Constant('false')]
        trinary = instructions.TrinaryOperator()

        result = trinary.fold(elems)

        self.assertEqual(len(result), 1)
        self.assertTrue(isinstance(result[0], instructions.Instructions))
        self.assertEqual(result[0].instructions, tuple(elems + [trinary]))


class TestSetOperator(tests.TestCase):
    def test_init(self):
        set_op = instructions.SetOperator(5)

        self.assertEqual(set_op.count, 5)
        self.assertEqual(set_op.opstr, 'set')

    def test_op(self):
        set_op = instructions.SetOperator(5)

        result = set_op.op(3, 5, 7)

        self.assertEqual(result, frozenset([3, 5, 7]))


class TestCallOperator(tests.TestCase):
    def test_init(self):
        call_op = instructions.CallOperator(5)

        self.assertEqual(call_op.count, 5)

    def test_repr(self):
        call_op = instructions.CallOperator(5)

        self.assertEqual(repr(call_op), "CallOperator(5)")

    def test_call_basic(self):
        func = mock.Mock(return_value='value', spec=[])
        ctxt = mock.Mock(stack=[func, 1, 2, 3, 4])
        call_op = instructions.CallOperator(5)

        call_op(ctxt)

        self.assertEqual(ctxt.stack, ['value'])
        func.assert_called_once_with(1, 2, 3, 4)

    def test_call_want_context_false(self):
        func = mock.Mock(return_value='value', _policies_want_context=False)
        ctxt = mock.Mock(stack=[func, 1, 2, 3, 4])
        call_op = instructions.CallOperator(5)

        call_op(ctxt)

        self.assertEqual(ctxt.stack, ['value'])
        func.assert_called_once_with(1, 2, 3, 4)

    def test_call_want_context_true(self):
        func = mock.Mock(return_value='value', _policies_want_context=True)
        ctxt = mock.Mock(stack=[func, 1, 2, 3, 4])
        call_op = instructions.CallOperator(5)

        call_op(ctxt)

        self.assertEqual(ctxt.stack, [])
        func.assert_called_once_with(ctxt, 1, 2, 3, 4)

    def test_hash(self):
        call_op = instructions.CallOperator(5)

        self.assertEqual(hash(call_op),
                         hash((instructions.CallOperator, 5)))

    def test_eq(self):
        class CallOperator2(instructions.CallOperator):
            pass

        call_op1 = instructions.CallOperator(5)
        call_op2 = instructions.CallOperator(5)
        call_op3 = instructions.CallOperator(3)
        call_op4 = CallOperator2(5)

        self.assertTrue(call_op1.__eq__(call_op2))
        self.assertFalse(call_op1.__eq__(call_op3))
        self.assertFalse(call_op1.__eq__(call_op4))

    # def test_op(self):
    #     func = mock.Mock(return_value='value')
    #     call_op = instructions.CallOperator(5)

    #     result = call_op.op(func, 1, 2, 3, 4)

    #     self.assertEqual(result, 'value')
    #     func.assert_called_once_with(1, 2, 3, 4)


class TestSetAuthorization(tests.TestCase):
    def test_repr(self):
        authz = instructions.SetAuthorization()

        self.assertEqual(repr(authz), "SetAuthorization()")

    @mock.patch('policies.authorization.Authorization', return_value='authz')
    def test_call(self, mock_Authorization):
        ctxt = mock.Mock(stack=['zhtua'], attrs='defaults')
        authz = instructions.SetAuthorization()

        authz(ctxt)

        self.assertEqual(ctxt.stack, [])
        self.assertEqual(ctxt.authz, 'authz')
        mock_Authorization.assert_called_once_with('zhtua', 'defaults')

    def test_hash(self):
        authz = instructions.SetAuthorization()

        self.assertEqual(hash(authz),
                         hash((instructions.SetAuthorization,)))

    def test_eq(self):
        class SetAuthorization2(instructions.SetAuthorization):
            pass

        authz1 = instructions.SetAuthorization()
        authz2 = instructions.SetAuthorization()
        authz3 = SetAuthorization2()

        self.assertTrue(authz1.__eq__(authz2))
        self.assertFalse(authz1.__eq__(authz3))


class TestAuthorizationAttr(tests.TestCase):
    def test_init(self):
        attr = instructions.AuthorizationAttr('attr')

        self.assertEqual(attr.attribute, 'attr')

    def test_repr(self):
        attr = instructions.AuthorizationAttr('attr')

        self.assertEqual(repr(attr), "AuthorizationAttr('attr')")

    def test_call(self):
        ctxt = mock.Mock(stack=['value'], authz=mock.Mock(_attrs={}))
        attr = instructions.AuthorizationAttr('attr')

        attr(ctxt)

        self.assertEqual(ctxt.stack, [])
        self.assertEqual(ctxt.authz._attrs, {'attr': 'value'})

    def test_hash(self):
        attr = instructions.AuthorizationAttr('attr')

        self.assertEqual(hash(attr),
                         hash((instructions.AuthorizationAttr, 'attr')))

    def test_eq(self):
        class AuthorizationAttr2(instructions.AuthorizationAttr):
            pass

        attr1 = instructions.AuthorizationAttr('attr')
        attr2 = instructions.AuthorizationAttr('attr')
        attr3 = instructions.AuthorizationAttr('other')
        attr4 = AuthorizationAttr2('attr')

        self.assertTrue(attr1.__eq__(attr2))
        self.assertFalse(attr1.__eq__(attr3))
        self.assertFalse(attr1.__eq__(attr4))


class TestInOperator(tests.TestCase):
    def test_op(self):
        exemplar = frozenset([1, 3, 5])

        self.assertTrue(instructions.in_op.op(1, exemplar))
        self.assertFalse(instructions.in_op.op(2, exemplar))
        self.assertTrue(instructions.in_op.op(3, exemplar))
        self.assertFalse(instructions.in_op.op(4, exemplar))
        self.assertTrue(instructions.in_op.op(5, exemplar))


class TestNotInOperator(tests.TestCase):
    def test_op(self):
        exemplar = frozenset([1, 3, 5])

        self.assertFalse(instructions.not_in_op.op(1, exemplar))
        self.assertTrue(instructions.not_in_op.op(2, exemplar))
        self.assertFalse(instructions.not_in_op.op(3, exemplar))
        self.assertTrue(instructions.not_in_op.op(4, exemplar))
        self.assertFalse(instructions.not_in_op.op(5, exemplar))


class TestAndOperator(tests.TestCase):
    def test_op(self):
        self.assertEqual(instructions.and_op.op(False, False), False)
        self.assertEqual(instructions.and_op.op(True, False), False)
        self.assertEqual(instructions.and_op.op(False, True), False)
        self.assertEqual(instructions.and_op.op(True, 'foo'), 'foo')


class TestOrOperator(tests.TestCase):
    def test_op(self):
        self.assertEqual(instructions.or_op.op(False, False), False)
        self.assertEqual(instructions.or_op.op('foo', False), 'foo')
        self.assertEqual(instructions.or_op.op(False, 'foo'), 'foo')
        self.assertEqual(instructions.or_op.op('foo', 'bar'), 'foo')


class TestItemOperator(tests.TestCase):
    def test_op(self):
        exemplar = {'a': 1, 'b': 2}

        self.assertEqual(instructions.item_op.op(exemplar, 'a'), 1)
        self.assertEqual(instructions.item_op.op(exemplar, 'b'), 2)
        self.assertRaises(KeyError, instructions.item_op.op,
                          exemplar, 'c')



# Copyright (C) 2013 by Kevin L. Mitchell <klmitch@mit.edu>
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

from policies import authorization

import tests


class TestAuthorization(tests.TestCase):
    def test_init_true(self):
        authz = authorization.Authorization(1)

        self.assertEqual(authz._result, True)
        self.assertEqual(authz._attrs, {})

    def test_init_false(self):
        authz = authorization.Authorization('')

        self.assertEqual(authz._result, False)
        self.assertEqual(authz._attrs, {})

    def test_init_attrs_indep(self):
        attrs = {'a': 1, 'b': 2, 'c': 3}
        authz = authorization.Authorization('', attrs)

        attrs['d'] = 4

        self.assertEqual(authz._attrs, {'a': 1, 'b': 2, 'c': 3})

    def test_getattr(self):
        attrs = {'a': 1, 'b': 2, 'c': 3}
        authz = authorization.Authorization(True, attrs)

        self.assertEqual(authz.a, 1)
        self.assertEqual(authz.b, 2)
        self.assertEqual(authz.c, 3)
        self.assertEqual(authz.d, None)

        authz._attrs['d'] = 4

        self.assertEqual(authz.d, 4)

        self.assertRaises(AttributeError, getattr, authz, '_a')

    def test_setattr(self):
        authz = authorization.Authorization(True, {})

        self.assertRaises(AttributeError, setattr, authz, 'a', 1)

    def test_delattr(self):
        def del_attr(az):
            del az.a

        authz = authorization.Authorization(True, {})

        self.assertRaises(AttributeError, delattr, authz, 'a')

    def test_boolean_true(self):
        authz = authorization.Authorization(True, {})

        self.assertTrue(authz)

    def test_boolean_false(self):
        authz = authorization.Authorization(False, {})

        self.assertFalse(authz)



# Copyright (C) 2013 by Kevin L. Mitchell <klmitch@mit.edu>
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.


class Authorization(object):
    """
    The authorization result.  Evaluates as either ``True`` or
    ``False``, depending on what the result of the policy evaluation
    was.  Can also have attributes set by the policy rule, which will
    default to ``None`` unless another default is provided.
    """

    def __init__(self, result, defaults=None):
        """
        Initialize an ``Authorization`` object.

        :param result: The result of the authorization evaluation.
                       Will be converted to boolean using the
                       ``bool()`` function.
        :param defaults: A dictionary of default values for the
                         authorization attributes.  If not provided,
                         authorization attributes default to ``None``.
        """

        super(Authorization, self).__setattr__('_result', bool(result))
        super(Authorization, self).__setattr__('_attrs', defaults.copy()
                                               if defaults else {})

    def __getattr__(self, name):
        """
        Retrieve a named authorization attribute.  If the
        authorization attribute has not been set, returns ``None``.

        :param name: The name of the authorization attribute to
                     return.

        :returns: The value of the authorization attribute.
        """

        # Prohibit access to internal attributes, so users don't get
        # any ideas...
        if name[0] == '_':
            raise AttributeError("%r object has no attribute %r" %
                                 (self.__class__.__name__, name))

        return self._attrs.get(name)

    def __setattr__(self, name, value):
        """
        Prohibits setting authorization attributes.  Authorization
        attributes are immutable.

        :param name: The name of the authorization attribute to set.
        :param value: The value to set the authorization attribute to.
        """

        raise AttributeError("%r object does not allow setting attribute %r" %
                             (self.__class__.__name__, name))

    def __delattr__(self, name):
        """
        Prohibits deletion of authorization attributes.  Authorization
        attributes are immutable.

        :param name: The name of the authorization attribute to set.
        """

        raise AttributeError("%r object does not allow deleting attribute %r" %
                             (self.__class__.__name__, name))

    def __nonzero__(self):
        """
        Returns the result of the authorization check.  If the
        ``result`` passed to the constructor evaluated as ``True``,
        returns ``True``; otherwise, returns ``False``.

        :returns: The result of the authorization check.
        """

        return self._result
    __bool__ = __nonzero__



# Copyright (C) 2013 by Kevin L. Mitchell <klmitch@mit.edu>
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

from policies import parser


class Rule(object):
    """
    Describe one policy rule.  A policy rule has a name and some text
    in the policy language, along with optional default values for
    authorization attributes.  The instructions for the rule text can
    be retrieved with the ``instructions`` property, and the rule text
    is parsed on demand.
    """

    def __init__(self, name, text='', attrs=None):
        """
        Initializes a ``Rule`` object.

        :param name: The rule name.
        :param text: Optional; the text for the rule.  If not given,
                     will be the empty string, which evaluates to a
                     rule which always denies authorization.
        :param attrs: A dictionary of authorization attribute
                      defaults.  Overrides the default of ``None``.
                      Note that authorization attributes may not have
                      a leading underscore ("_").
        """

        # Store the essential data
        self.name = name
        self.text = text
        self.attrs = dict((k, v) for k, v in (attrs or {}).items()
                          if k[0] != '_' and v is not None)

        # The instructions will be parsed on demand
        self._instructions = None

    @property
    def instructions(self):
        """
        Retrieve the instructions for the rule.
        """

        if self._instructions is None:
            # Compile the rule into an Instructions instance; we do
            # this lazily to amortize the cost of the compilation,
            # then cache that result for efficiency...
            self._instructions = parser.parse_rule(self.name, self.text)

        return self._instructions


class RuleDoc(object):
    """
    A simple class to store documentation for a rule.  Instances of
    this class are constructed when pre-declaring rules on a given
    ``RuleSet``.
    """

    def __init__(self, name, doc=None, attr_docs=None):
        """
        Initialize a ``RuleDoc`` object.

        :param name: The name of the rule.
        :param doc: A documentation string for the rule.  If given,
                    should describe the purpose of the rule--that is,
                    what operation is protected by the rule.
        :param attr_docs: A dictionary of documentation strings for
                          the authorization attributes.  If provided,
                          each entry should explain the purpose of
                          each authorization attribute--that is, what
                          operation is protected or goverened by the
                          attribute.
        """

        self.name = name
        self.doc = doc

        # Filter out invalid authorization attributes (those beginning
        # with an '_') and those that have no documentation
        self.attr_docs = dict((k, v) for k, v in (attr_docs or {}).items()
                              if k[0] != '_' and v)



# Copyright (C) 2013 by Kevin L. Mitchell <klmitch@mit.edu>
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

import policies

import tests


# Need a user class for our tests
class User(object):
    def __init__(self, name, groups=None, admin=False):
        self.name = name
        self.groups = set(groups or [])
        self.admin = admin

    def __eq__(self, other):
        return self.name == other.name

    def in_group(self, group):
        return group in self.groups


class TestRules(tests.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.policy = policies.Policy()
        cls.policy['is_admin'] = """
            user.in_group("administrators") and user.admin
        """
        cls.policy['user_update'] = """
            user == target or rule("is_admin") {{
                payment=rule("is_admin"),
                name=user == target,
            }}
        """

        cls.alice = User('alice')
        cls.bob = User('bob')
        cls.charlie = User('charlie', ['administrators'])
        cls.charlie_admin = User('charlie', ['administrators'], True)
        cls.deborah = User('deborah', [], True)

    def evaluate(self, user, target):
        return self.policy.evaluate('user_update',
                                    {'user': user, 'target': target})

    def test_same_user(self):
        result = self.evaluate(self.alice, self.alice)

        self.assertTrue(result)
        self.assertFalse(result.payment)
        self.assertTrue(result.name)

    def test_other_user(self):
        result = self.evaluate(self.alice, self.bob)

        self.assertFalse(result)
        self.assertFalse(result.payment)
        self.assertFalse(result.name)

    def test_admin_user_only(self):
        result = self.evaluate(self.charlie, self.bob)

        self.assertFalse(result)
        self.assertFalse(result.payment)
        self.assertFalse(result.name)

    def test_user_privileges(self):
        result = self.evaluate(self.deborah, self.bob)

        self.assertFalse(result)
        self.assertFalse(result.payment)
        self.assertFalse(result.name)

    def test_admin_privileges(self):
        result = self.evaluate(self.charlie_admin, self.bob)

        self.assertTrue(result)
        self.assertTrue(result.payment)
        self.assertFalse(result.name)

    def test_no_rule(self):
        result = self.policy.evaluate('no_such',
                                      {'user': self.alice, 'target': self.bob})

        self.assertFalse(result)
        self.assertEqual(result.payment, None)
        self.assertEqual(result.name, None)



# Copyright (C) 2013 by Kevin L. Mitchell <klmitch@mit.edu>
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

import mock

from policies import rules

import tests


class TestRule(tests.TestCase):
    def test_init_basic(self):
        rule = rules.Rule('name')

        self.assertEqual(rule.name, 'name')
        self.assertEqual(rule.text, '')
        self.assertEqual(rule.attrs, {})
        self.assertEqual(rule._instructions, None)

    def test_init_full(self):
        rule = rules.Rule('name', 'text', {
            'name': 'eman',
            'text': 'txet',
            'other': 1,
            '_ignored': 'ignored',
            'none': None,
        })

        self.assertEqual(rule.name, 'name')
        self.assertEqual(rule.text, 'text')
        self.assertEqual(rule.attrs, {
            'name': 'eman',
            'text': 'txet',
            'other': 1,
        })
        self.assertEqual(rule._instructions, None)

    @mock.patch('policies.parser.parse_rule', return_value='instructions')
    def test_instructions_cached(self, mock_parse_rule):
        rule = rules.Rule('name', 'text')
        rule._instructions = 'cached'

        self.assertEqual(rule.instructions, 'cached')
        self.assertEqual(rule._instructions, 'cached')
        self.assertFalse(mock_parse_rule.called)

    @mock.patch('policies.parser.parse_rule', return_value='instructions')
    def test_instructions_uncached(self, mock_parse_rule):
        rule = rules.Rule('name', 'text')

        self.assertEqual(rule.instructions, 'instructions')
        self.assertEqual(rule._instructions, 'instructions')
        mock_parse_rule.assert_called_once_with('name', 'text')


class TestRuleDoc(tests.TestCase):
    def test_init_basic(self):
        rdoc = rules.RuleDoc('name')

        self.assertEqual(rdoc.name, 'name')
        self.assertEqual(rdoc.doc, None)
        self.assertEqual(rdoc.attr_docs, {})

    def test_init_full(self):
        rdoc = rules.RuleDoc('name', 'doc', {
            'name': 'eman',
            'text': 'txet',
            '_ignored': 'ignored',
            'none': '',
        })

        self.assertEqual(rdoc.name, 'name')
        self.assertEqual(rdoc.doc, 'doc')
        self.assertEqual(rdoc.attr_docs, {
            'name': 'eman',
            'text': 'txet',
        })



# Copyright (C) 2013 by Kevin L. Mitchell <klmitch@mit.edu>
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

import pyparsing
import mock

from policies.instructions import *
from policies import parser

import tests


class TestUnaryConstruct(tests.TestCase):
    def test_call(self):
        result = parser.unary_construct([inv_op, Ident('a')])

        self.assertEqual(result, [Instructions([Ident('a'), inv_op])])


class TestBinaryConstruct(tests.TestCase):
    def test_one_sequence(self):
        result = parser.binary_construct([Ident('a'), add_op, Constant(2)])

        self.assertEqual(result, [Instructions([
            Ident('a'), Constant(2), add_op,
        ])])

    def test_multi_sequence(self):
        result = parser.binary_construct([
            Ident('a'), add_op, Constant(2), add_op, Ident('b'), add_op,
            Constant(3),
        ])

        self.assertEqual(result, [Instructions([
            Ident('a'), Constant(2), add_op, Ident('b'), add_op, Constant(3),
            add_op,
        ])])


class TestParseRule(tests.TestCase):
    @mock.patch('logging.getLogger')
    @mock.patch.object(parser.rule, 'parseString', return_value=['success'])
    def test_success(self, mock_parseString, mock_getLogger):
        result = parser.parse_rule('test', 'rule text')

        self.assertEqual(result, 'success')
        mock_parseString.assert_called_once_with('rule text', parseAll=True)
        self.assertFalse(mock_getLogger.called)

    @mock.patch('logging.getLogger')
    @mock.patch.object(parser.rule, 'parseString',
                       side_effect=pyparsing.ParseException(
                           "test rule string", loc=5, msg="trial error"))
    def test_failure(self, mock_parseString, mock_getLogger):
        result = parser.parse_rule('test', 'rule text')

        self.assertEqual(result, Instructions([Constant(False), set_authz]))
        mock_parseString.assert_called_once_with('rule text', parseAll=True)
        mock_getLogger.assert_called_once_with('policies')
        mock_getLogger.return_value.assert_has_calls([
            mock.call.warn("Failed to parse rule 'test': "
                           "trial error (at char 5), (line:1, col:6)"),
            mock.call.warn("Rule line: test rule string"),
            mock.call.warn("Location :      ^"),
        ])

    @mock.patch('logging.getLogger')
    @mock.patch.object(parser.rule, 'parseString',
                       side_effect=pyparsing.ParseException(
                           "test rule string", loc=5, msg="trial error"))
    def test_raise(self, mock_parseString, mock_getLogger):
        self.assertRaises(pyparsing.ParseException,
                          parser.parse_rule, 'test', 'rule text',
                          do_raise=True)

        mock_parseString.assert_called_once_with('rule text', parseAll=True)
        self.assertFalse(mock_getLogger.called)

    rules = [
        ('True', Instructions([Constant(True), set_authz])),
        ('False', Instructions([Constant(False), set_authz])),
        ('None', Instructions([Constant(None), set_authz])),
        ('23', Instructions([Constant(23), set_authz])),
        ('-23', Instructions([Constant(-23), set_authz])),
        ('23.', Instructions([Constant(23.0), set_authz])),
        ('23.1', Instructions([Constant(23.1), set_authz])),
        ('-23e1', Instructions([Constant(-230.0), set_authz])),
        ('-23e+1', Instructions([Constant(-230.0), set_authz])),
        ('23e-1', Instructions([Constant(2.3), set_authz])),
        ('"this is \\" a test"', Instructions([
            Constant('this is " a test'), set_authz,
        ])),
        ('"string " "concatenation"', Instructions([
            Constant('string concatenation'), set_authz,
        ])),
        ('foobar', Instructions([Ident('foobar'), set_authz])),
        ('~a', Instructions([Ident('a'), inv_op, set_authz])),
        ('+a', Instructions([Ident('a'), pos_op, set_authz])),
        ('-a', Instructions([Ident('a'), neg_op, set_authz])),
        ('not a', Instructions([Ident('a'), not_op, set_authz])),
        ('a ** b', Instructions([Ident('a'), Ident('b'), pow_op, set_authz])),
        ('a * b', Instructions([Ident('a'), Ident('b'), mul_op, set_authz])),
        ('a / b', Instructions([
            Ident('a'), Ident('b'), true_div_op, set_authz,
        ])),
        ('a // b', Instructions([
            Ident('a'), Ident('b'), floor_div_op, set_authz,
        ])),
        ('a % b', Instructions([Ident('a'), Ident('b'), mod_op, set_authz])),
        ('a + b', Instructions([Ident('a'), Ident('b'), add_op, set_authz])),
        ('a - b', Instructions([Ident('a'), Ident('b'), sub_op, set_authz])),
        ('a + -2', Instructions([
            Ident('a'), Constant(-2), add_op, set_authz,
        ])),
        ('a << b', Instructions([
            Ident('a'), Ident('b'), left_shift_op, set_authz,
        ])),
        ('a >> b', Instructions([
            Ident('a'), Ident('b'), right_shift_op, set_authz,
        ])),
        ('a & b', Instructions([
            Ident('a'), Ident('b'), bit_and_op, set_authz,
        ])),
        ('a ^ b', Instructions([
            Ident('a'), Ident('b'), bit_xor_op, set_authz,
        ])),
        ('a | b', Instructions([
            Ident('a'), Ident('b'), bit_or_op, set_authz,
        ])),
        ('a in b', Instructions([Ident('a'), Ident('b'), in_op, set_authz])),
        ('a not in b', Instructions([
            Ident('a'), Ident('b'), not_in_op, set_authz,
        ])),
        ('a is b', Instructions([Ident('a'), Ident('b'), is_op, set_authz])),
        ('a is not b', Instructions([
            Ident('a'), Ident('b'), is_not_op, set_authz,
        ])),
        ('a < b', Instructions([Ident('a'), Ident('b'), lt_op, set_authz])),
        ('a > b', Instructions([Ident('a'), Ident('b'), gt_op, set_authz])),
        ('a <= b', Instructions([Ident('a'), Ident('b'), le_op, set_authz])),
        ('a >= b', Instructions([Ident('a'), Ident('b'), ge_op, set_authz])),
        ('a != b', Instructions([Ident('a'), Ident('b'), ne_op, set_authz])),
        ('a == b', Instructions([Ident('a'), Ident('b'), eq_op, set_authz])),
        ('a and b', Instructions([Ident('a'), Ident('b'), and_op, set_authz])),
        ('a or b', Instructions([Ident('a'), Ident('b'), or_op, set_authz])),
        ('a if b else c', Instructions([
            Ident('b'), Ident('a'), Ident('c'), trinary_op, set_authz
        ])),
        ('a + b * c + d', Instructions([
            Ident('a'), Ident('b'), Ident('c'), mul_op, add_op, Ident('d'),
            add_op, set_authz,
        ])),
        ('a * b + c * d', Instructions([
            Ident('a'), Ident('b'), mul_op, Ident('c'), Ident('d'), mul_op,
            add_op, set_authz,
        ])),
        ('{}', Instructions([Constant(frozenset()), set_authz])),
        ('{,}', Instructions([Constant(frozenset()), set_authz])),
        ('{a}', Instructions([Ident('a'), SetOperator(1), set_authz])),
        ('{a, b, c}', Instructions([
            Ident('a'), Ident('b'), Ident('c'), SetOperator(3), set_authz,
        ])),
        ('{a, b, c,}', Instructions([
            Ident('a'), Ident('b'), Ident('c'), SetOperator(3), set_authz,
        ])),
        ('a.b', Instructions([Ident('a'), Attribute('b'), set_authz])),
        ('a[3]', Instructions([Ident('a'), Constant(3), item_op, set_authz])),
        ('a()', Instructions([Ident('a'), CallOperator(1), set_authz])),
        ('a(,)', Instructions([Ident('a'), CallOperator(1), set_authz])),
        ('a(1)', Instructions([
            Ident('a'), Constant(1), CallOperator(2), set_authz,
        ])),
        ('a(1, 2, 3)', Instructions([
            Ident('a'), Constant(1), Constant(2), Constant(3), CallOperator(4),
            set_authz,
        ])),
        ('a(1, 2, 3,)', Instructions([
            Ident('a'), Constant(1), Constant(2), Constant(3), CallOperator(4),
            set_authz,
        ])),
        ('True {{}}', Instructions([Constant(True), set_authz])),
        ('True {{,}}', Instructions([Constant(True), set_authz])),
        ('True {{a=1}}', Instructions([
            Constant(True), set_authz,
            Constant(1), AuthorizationAttr('a'),
        ])),
        ('True {{a=1,b=2,c=3}}', Instructions([
            Constant(True), set_authz,
            Constant(1), AuthorizationAttr('a'),
            Constant(2), AuthorizationAttr('b'),
            Constant(3), AuthorizationAttr('c'),
        ])),
        ('True {{a=1,b=2,c=3,}}', Instructions([
            Constant(True), set_authz,
            Constant(1), AuthorizationAttr('a'),
            Constant(2), AuthorizationAttr('b'),
            Constant(3), AuthorizationAttr('c'),
        ])),
        ('level > 400 {{ level=level }}', Instructions([
            Ident('level'), Constant(400), gt_op, set_authz,
            Ident('level'), AuthorizationAttr('level'),
        ])),
    ]

    def test_parse(self):
        errors = 0
        for text, expected in self.rules:
            try:
                result = parser.parse_rule("test", text, do_raise=True)
            except pyparsing.ParseException as exc:
                if expected is not None:
                    # Print out a description of the unexpected failure
                    print('')
                    print("Failure to parse %r: %s" % (text, exc))
                    print("  Line    : %s" % exc.line)
                    print("  Location: %s^" % (" " * (exc.col - 1)))
                    errors += 1
                continue

            # Compare the expected to the actual
            if result != expected:
                print('')
                print("Failure to parse %r: %r != %r" %
                      (text, result, expected))
                errors += 1

        if errors > 0:
            self.fail("Parse failures encountered; see output for information")



# Copyright (C) 2013 by Kevin L. Mitchell <klmitch@mit.edu>
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

import collections
import contextlib
import logging
import sys

import pkg_resources
import six

from policies import authorization
from policies import rules


class PolicyException(Exception):
    """
    An exception raised if an attempt is made to store a given rule in
    the wrong location in a ``Policy``.
    """

    pass


class PolicyContext(object):
    """
    A context object for evaluating authorization rules.  Contains a
    ``stack``--a simple list onto which values will be pushed and from
    which values will be popped--as well as providing storage for the
    final authorization object, ``authz``.  Also stores the ``Policy``
    object (``policy``), the ``variables`` for the evaluation, and
    ``attrs``, a dictionary of default values for authorization
    attributes.
    """

    def __init__(self, policy, attrs, variables):
        """
        Initialize a ``PolicyContext`` object.

        :param policy: The ``Policy`` object.
        :param attrs: A dictionary of authorization attribute default
                      values.
        :param variables: A dictionary of variables to be defined for
                          the evaluation.
        """

        # Save parameters
        self.policy = policy
        self.attrs = attrs
        self.variables = variables

        # Set up the stack
        self._name = []
        self.stack = []
        self.authz = None

        # Add a cache for rules
        self.rule_cache = {}

        # Used to keep track of error reporting, to ensure that an
        # exception raised at one level of nesting isn't reported
        # multiple times
        self.reported = False

    def resolve(self, symbol):
        """
        Resolve a symbol encountered during a rule evaluation into the
        actual value for that symbol.

        :param symbol: The symbol being resolved.

        :returns: The value of that symbol.  If the symbol was not
                  declared in the ``variables`` parameter of the
                  constructor, a call will be made to the ``Policy``'s
                  ``resolve()`` method.
        """

        # Try the variables first
        if symbol in self.variables:
            return self.variables[symbol]

        return self.policy.resolve(symbol)

    @contextlib.contextmanager
    def push_rule(self, name):
        """
        Allow one rule to be evaluated in the context of another.
        This allows keeping track of the rule names during nested rule
        evaluation.

        :param name: The name of the nested rule to be evaluated.

        :returns: A context manager, suitable for use with the
                  ``with`` statement.  No value is generated.
        """

        # Verify that we haven't been evaluating the rule already;
        # this is to prohibit recursive rules from locking us up...
        if name in self._name:
            raise PolicyException(
                "Rule recursion detected; invocation chain: %s -> %s" %
                (' -> '.join(self._name), name))

        # Save the name temporarily
        self._name.append(name)
        try:
            yield
        except Exception as exc:
            exc_info = sys.exc_info()

            # Report only if we haven't reported it yet
            if not self.reported:
                # Get the logger and emit a log message
                log = logging.getLogger('policies')
                log.warn("Exception raised while evaluating rule %r: %s" %
                         (name, exc))
                self.reported = True

            six.reraise(*exc_info)
        finally:
            # Pop the name off the stack of names
            self._name.pop()

    @property
    def name(self):
        """
        Retrieve the name of the rule being evaluated.
        """

        return self._name[-1] if self._name else None


class Policy(collections.MutableMapping):
    """
    Maintain a mapping of rules, along with declared defaults and any
    documentation information.  Allows for access to any rule, as well
    as evaluation of a rule.
    """

    # Pre-populate the resolver cache with these special callables
    builtins = {
        'abs': abs,
        'basestring': six.string_types,
        'bin': bin,
        'bool': bool,
        'bytes': six.binary_type,
        'callable': callable,
        'chr': chr,
        'complex': complex,
        'dict': dict,
        'divmod': divmod,
        'enumerate': enumerate,
        'float': float,
        'format': format,
        'frozenset': frozenset,
        'getattr': getattr,
        'hasattr': hasattr,
        'hash': hash,
        'hex': hex,
        'id': id,
        'int': int,
        'isinstance': isinstance,
        'issubclass': issubclass,
        'iter': iter,
        'len': len,
        'list': list,
        'long': long if not six.PY3 else int,
        'max': max,
        'min': min,
        'next': next,
        'object': object,
        'oct': oct,
        'ord': ord,
        'pow': pow,
        'range': range,
        'repr': repr,
        'reversed': reversed,
        'round': round,
        'set': set,
        'sorted': sorted,
        'str': six.text_type,
        'sum': sum,
        'tuple': tuple,
        'type': type,
        'unichr': six.unichr,
        'unicode': six.text_type,
        'xrange': range,
        'zip': zip,
    }

    # The context class to use
    context_class = PolicyContext

    def __init__(self, group=None, builtins=None):
        """
        Initialize a ``Policy`` object.

        :param group: An entrypoint group to search for unresolved
                      symbols.  This allows for extension of the
                      policy rules for a given application.  If not
                      given, unresolved symbols will resolve to
                      ``None``.
        :param builtins: A dictionary of "builtin" functions, which
                         overrides any symbols defined in the
                         entrypoint group.  If not provided, a default
                         of select Python builtins will be used
                         instead.
        """

        # Save the entrypoint group
        self._group = group

        # Set up the mappings
        self._defaults = {}
        self._docs = {}
        self._rules = {}

        # Seed the resolve cache
        if builtins is None:
            builtins = self.builtins
        self._resolve_cache = builtins.copy()

        # Add the default rule
        self._resolve_cache.setdefault('rule', rule)

    def __getitem__(self, key):
        """
        Retrieve a ``Rule`` given its name.  Raises a ``KeyError`` if
        the rule is both undefined and undeclared.

        :param key: The name of the rule to get.

        :returns: The ``Rule`` object describing the rule.
        """

        # Check to see if the rule has been set
        if key in self._rules:
            return self._rules[key]

        # If it's been declared, return the default
        if key in self._defaults:
            return self._defaults[key]

        raise KeyError(key)

    def __setitem__(self, key, rule):
        """
        Set a ``Rule`` with a given name.  Raises a
        ``PolicyException`` if the key and the name of the rule don't
        match.

        :param key: The name of the rule to set.
        :param rule: Either a ``Rule`` object with a name matching
                     ``key``, or the text of the rule.
        """

        if isinstance(rule, six.string_types):
            # Construct the rule from the string
            rule = rules.Rule(key, rule)
        elif key != rule.name:
            raise PolicyException("key %r does not match rule name %r" %
                                  (key, rule.name))

        self._rules[key] = rule

    def __delitem__(self, key):
        """
        Reset a rule with a given name to its default.

        :param key: The name of the rule to reset.
        """

        del self._rules[key]

    def __iter__(self):
        """
        Iterate over the rule names.

        :returns: An iterator over the rule names.
        """

        return iter(set(self._defaults.keys()) | set(self._rules.keys()))

    def __len__(self):
        """
        Obtain the number of rules available on the ``Policy``.

        :returns: The number of independent rules on the ``Policy``.
        """

        return len(set(self._defaults.keys()) | set(self._rules.keys()))

    def declare(self, name, text='', doc=None, attrs=None, attr_docs=None):
        """
        Declare a rule.  This allows a default for a given rule to be
        set, along with default values for the authorization
        attributes.  This function can also include documentation for
        the rule and the authorization attributes, allowing a sample
        policy configuration file to be generated.

        :param name: The name of the rule.
        :param text: The text of the rule.  Defaults to the empty
                     string.
        :param doc: A string documenting the purpose of the rule.
        :param attrs: A dictionary of default values for the
                      authorization attributes.  Note that
                      authorization attributes cannot have names
                      beginning with an underscore ("_").
        :param attr_docs: A dictionary of strings for documenting the
                          purpose of the authorization attributes.
        """

        self._defaults[name] = rules.Rule(name, text, attrs)
        self._docs[name] = rules.RuleDoc(name, doc, attr_docs)

        return self._defaults[name]

    def set_rule(self, rule):
        """
        Stores a ``Rule`` object in the ``Policy``.

        :param rule: The rule to save.
        """

        self._rules[rule.name] = rule

    def del_rule(self, rule):
        """
        Deletes a ``Rule`` object from the ``Policy``.  This restores
        the rule to the default set using ``declare()``.

        :param rule: The rule to delete.
        """

        del self._rules[rule.name]

    def get_doc(self, name):
        """
        Retrieve a ``RuleDoc`` object from the ``Policy`` with the
        given name.  The ``RuleDoc`` object contains all documentation
        for the named rule.

        :param name: The name of the rule to retrieve the
                     documentation for.

        :returns: A ``RuleDoc`` object containing the documentation
                  for the rule.
        """

        # Create one if there isn't one already
        if name not in self._docs:
            self._docs[name] = rules.RuleDoc(name)

        return self._docs[name]

    def get_docs(self):
        """
        Retrieve all declared ``RuleDoc`` objects from the ``Policy``.
        The ``RuleDoc`` object contains all documentation for the
        declared rules.

        :returns: A list of ``RuleDoc`` objects containing
                  documentation for the declared rules.
        """

        return self._docs.values()

    def resolve(self, symbol):
        """
        Resolve a symbol using the entrypoint group.

        :param symbol: The symbol being resolved.

        :returns: The value of that symbol.  If the symbol cannot be
                  found, or if no entrypoint group was passed to the
                  constructor, will return ``None``.
        """

        # Search for a corresponding symbol
        if symbol not in self._resolve_cache:
            result = None

            # Search through entrypoints only if we have a group
            if self._group is not None:
                for ep in pkg_resources.iter_entry_points(self._group, symbol):
                    try:
                        result = ep.load()
                    except (ImportError, AttributeError,
                            pkg_resources.UnknownExtra):
                        continue

                    # We found the result we were looking for
                    break

            # Cache the result
            self._resolve_cache[symbol] = result

        return self._resolve_cache[symbol]

    def evaluate(self, name, variables=None):
        """
        Evaluate a named rule.

        :param name: The name of the rule to evaluate.
        :param variables: An optional dictionary of variables to make
                          available during evaluation of the rule.

        :returns: An instance of
                  ``policies.authorization.Authorization`` with the
                  result of the rule evaluation.  This will include
                  any authorization attributes.
        """

        # Get the rule and predeclaration
        rule = self._rules.get(name)
        default = self._defaults.get(name)

        # Short-circuit if we don't have either
        if rule is None and default is None:
            return authorization.Authorization(False)

        # Marry the attribute defaults
        attrs = {}
        if default:
            attrs.update(default.attrs)
        if rule:
            attrs.update(rule.attrs)

        # Select the rule we'll actually use
        if rule is None:
            rule = default

        # Construct the context
        ctxt = self.context_class(self, attrs, variables or {})

        # Execute the rule
        try:
            with ctxt.push_rule(name):
                rule.instructions(ctxt)
        except Exception as exc:
            # Fail closed
            return authorization.Authorization(False, attrs)

        # Return the authorization result
        return ctxt.authz


def want_context(func):
    """
    A decorator that marks a policy function as wanting the evaluation
    context object.

    :param func: The function to be decorated.

    :returns: The decorated function.
    """

    func._policies_want_context = True

    return func


@want_context
def rule(ctxt, name):
    """
    Allows evaluation of another rule while evaluating a rule.

    :param ctxt: The evaluation context for the rule.
    :param name: The name of the rule to evaluate.
    """

    # If the result of evaluation is in the rule cache, bypass
    # evaluation
    if name in ctxt.rule_cache:
        ctxt.stack.append(ctxt.rule_cache[name])
        return

    # Obtain the rule we're to evaluate
    try:
        rule = ctxt.policy[name]
    except KeyError:
        # Rule doesn't exist; log a message and assume False
        log = logging.getLogger('policies')
        log.warn("Request to evaluate non-existant rule %r "
                 "while evaluating rule %r" % (name, ctxt.name))
        ctxt.stack.append(False)
        ctxt.rule_cache[name] = False
        return

    # Evaluate the rule, stopping at the set_authz instruction
    with ctxt.push_rule(name):
        rule.instructions(ctxt, True)

    # Cache the result
    ctxt.rule_cache[name] = ctxt.stack[-1]



# Copyright (C) 2013 by Kevin L. Mitchell <klmitch@mit.edu>
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

import mock
import pkg_resources

from policies import policy
from policies import rules

import tests


class TestPolicyContext(tests.TestCase):
    def test_init(self):
        ctxt = policy.PolicyContext('policy', 'attrs', 'variables')

        self.assertEqual(ctxt.policy, 'policy')
        self.assertEqual(ctxt.attrs, 'attrs')
        self.assertEqual(ctxt.variables, 'variables')
        self.assertEqual(ctxt._name, [])
        self.assertEqual(ctxt.stack, [])
        self.assertEqual(ctxt.authz, None)
        self.assertEqual(ctxt.rule_cache, {})
        self.assertEqual(ctxt.reported, False)

    def test_resolve_defined(self):
        pol = mock.Mock()
        ctxt = policy.PolicyContext(pol, 'attrs', {'a': 1})

        result = ctxt.resolve('a')

        self.assertEqual(result, 1)
        self.assertFalse(pol.resolve.called)

    def test_resolve_undefined(self):
        pol = mock.Mock(**{'resolve.return_value': 'value'})
        ctxt = policy.PolicyContext(pol, 'attrs', {'a': 1})

        result = ctxt.resolve('b')

        self.assertEqual(result, 'value')
        pol.resolve.assert_called_once_with('b')

    @mock.patch('logging.getLogger')
    def test_push_rule_success(self, mock_getLogger):
        ctxt = policy.PolicyContext('policy', 'attrs', 'variables')
        ctxt._name = ['rule1', 'rule2']

        with ctxt.push_rule('rule3'):
            self.assertEqual(ctxt._name, ['rule1', 'rule2', 'rule3'])

        self.assertEqual(ctxt._name, ['rule1', 'rule2'])
        self.assertEqual(ctxt.reported, False)
        self.assertFalse(mock_getLogger.called)

    @mock.patch('logging.getLogger')
    def test_push_rule_failure_reported(self, mock_getLogger):
        ctxt = policy.PolicyContext('policy', 'attrs', 'variables')
        ctxt._name = ['rule1', 'rule2']
        ctxt.reported = True

        try:
            with ctxt.push_rule('rule3'):
                raise tests.TestException('test')
        except tests.TestException:
            pass
        else:
            self.fail("TestException failed to bubble up")

        self.assertEqual(ctxt._name, ['rule1', 'rule2'])
        self.assertEqual(ctxt.reported, True)
        self.assertFalse(mock_getLogger.called)

    @mock.patch('logging.getLogger')
    def test_push_rule_failure_unreported(self, mock_getLogger):
        ctxt = policy.PolicyContext('policy', 'attrs', 'variables')
        ctxt._name = ['rule1', 'rule2']

        try:
            with ctxt.push_rule('rule3'):
                raise tests.TestException('test')
        except tests.TestException:
            pass
        else:
            self.fail("TestException failed to bubble up")

        self.assertEqual(ctxt._name, ['rule1', 'rule2'])
        self.assertEqual(ctxt.reported, True)
        mock_getLogger.assert_called_once_with('policies')
        mock_getLogger.return_value.warn.assert_called_once_with(
            "Exception raised while evaluating rule 'rule3': test")

    @mock.patch('logging.getLogger')
    def test_push_rule_recursion(self, mock_getLogger):
        ctxt = policy.PolicyContext('policy', 'attrs', 'variables')
        ctxt._name = ['rule1', 'rule2', 'rule3', 'rule4']

        try:
            with ctxt.push_rule('rule2'):
                pass
        except policy.PolicyException:
            pass
        else:
            self.fail("Failed to detect rule recursion")

        self.assertEqual(ctxt._name, ['rule1', 'rule2', 'rule3', 'rule4'])
        self.assertEqual(ctxt.reported, False)
        self.assertFalse(mock_getLogger.called)

    def test_name_defined(self):
        ctxt = policy.PolicyContext('policy', 'attrs', 'variables')
        ctxt._name = ['rule1', 'rule2']

        self.assertEqual(ctxt.name, 'rule2')

    def test_name_undefined(self):
        ctxt = policy.PolicyContext('policy', 'attrs', 'variables')

        self.assertEqual(ctxt.name, None)


def item_setter(obj, item, value):
    obj[item] = value


class TestPolicy(tests.TestCase):
    def test_init_basic(self):
        expected = policy.Policy.builtins.copy()
        expected['rule'] = policy.rule

        pol = policy.Policy()

        self.assertEqual(pol._group, None)
        self.assertEqual(pol._defaults, {})
        self.assertEqual(pol._docs, {})
        self.assertEqual(pol._rules, {})
        self.assertEqual(pol._resolve_cache, expected)
        self.assertNotEqual(id(pol._resolve_cache),
                            id(policy.Policy.builtins))

    def test_init_full(self):
        builtins = {'a': 1, 'b': 2, 'c': 3}
        expected = builtins.copy()
        expected['rule'] = policy.rule

        pol = policy.Policy('group', builtins)

        self.assertEqual(pol._group, 'group')
        self.assertEqual(pol._defaults, {})
        self.assertEqual(pol._docs, {})
        self.assertEqual(pol._rules, {})
        self.assertEqual(pol._resolve_cache, expected)
        self.assertNotEqual(id(pol._resolve_cache), id(builtins))

    def test_getitem_none(self):
        pol = policy.Policy()

        self.assertRaises(KeyError, pol.__getitem__, 'rule')

    def test_getitem_default(self):
        pol = policy.Policy()
        pol._defaults['rule'] = 'default'

        self.assertEqual(pol['rule'], 'default')

    def test_getitem_set(self):
        pol = policy.Policy()
        pol._defaults['rule'] = 'default'
        pol._rules['rule'] = 'configured'

        self.assertEqual(pol['rule'], 'configured')

    @mock.patch.object(rules, 'Rule', return_value='compiled')
    def test_setitem_string(self, mock_Rule):
        pol = policy.Policy()

        pol['rule'] = 'test'

        self.assertEqual(pol._rules, {'rule': 'compiled'})
        self.assertEqual(pol._defaults, {})
        mock_Rule.assert_called_once_with('rule', 'test')

    def test_setitem_badname(self):
        pol = policy.Policy()
        rule = rules.Rule('name')

        with mock.patch.object(rules, 'Rule',
                               return_value='compiled') as mock_Rule:
            self.assertRaises(policy.PolicyException, item_setter,
                              pol, 'rule', rule)

        self.assertEqual(pol._defaults, {})
        self.assertFalse(mock_Rule.called)

    def test_setitem(self):
        pol = policy.Policy()
        rule = rules.Rule('rule')

        with mock.patch.object(rules, 'Rule',
                               return_value='compiled') as mock_Rule:
            pol['rule'] = rule

        self.assertEqual(pol._rules, {'rule': rule})
        self.assertEqual(pol._defaults, {})
        self.assertFalse(mock_Rule.called)

    def test_delitem(self):
        pol = policy.Policy()
        pol._rules = {'a': 1, 'b': 2}
        pol._defaults = {'a': 3}

        del pol['a']

        self.assertEqual(pol._rules, {'b': 2})
        self.assertEqual(pol._defaults, {'a': 3})

    def test_iter(self):
        pol = policy.Policy()
        pol._rules = {'a': 1, 'c': 3}
        pol._defaults = {'b': 2, 'c': 3}

        result = sorted(iter(pol))

        self.assertEqual(result, ['a', 'b', 'c'])

    def test_len(self):
        pol = policy.Policy()
        pol._rules = {'a': 1, 'c': 3}
        pol._defaults = {'b': 2, 'c': 3}

        result = len(pol)

        self.assertEqual(result, 3)

    @mock.patch.object(rules, 'Rule', return_value='rule')
    @mock.patch.object(rules, 'RuleDoc', return_value='doc')
    def test_declare_basic(self, mock_RuleDoc, mock_Rule):
        pol = policy.Policy()

        pol.declare('name')

        self.assertEqual(pol._defaults, {'name': 'rule'})
        self.assertEqual(pol._docs, {'name': 'doc'})
        self.assertEqual(pol._rules, {})
        mock_Rule.assert_called_once_with('name', '', None)
        mock_RuleDoc.assert_called_once_with('name', None, None)

    @mock.patch.object(rules, 'Rule', return_value='rule')
    @mock.patch.object(rules, 'RuleDoc', return_value='doc')
    def test_declare_full(self, mock_RuleDoc, mock_Rule):
        pol = policy.Policy()

        pol.declare('name', 'text', 'doc', {'a': 1, 'b': 2},
                    {'a': 'doc a', 'b': 'doc b'})

        self.assertEqual(pol._defaults, {'name': 'rule'})
        self.assertEqual(pol._docs, {'name': 'doc'})
        self.assertEqual(pol._rules, {})
        mock_Rule.assert_called_once_with('name', 'text', {'a': 1, 'b': 2})
        mock_RuleDoc.assert_called_once_with('name', 'doc',
                                             {'a': 'doc a', 'b': 'doc b'})

    def test_set_rule(self):
        rule = rules.Rule('name')
        pol = policy.Policy()

        pol.set_rule(rule)

        self.assertEqual(pol._rules, {'name': rule})
        self.assertEqual(pol._defaults, {})

    def test_del_rule(self):
        rule = rules.Rule('name')
        pol = policy.Policy()
        pol._rules = {'name': 'rule1', 'other': 'rule2'}
        pol._defaults = {'name': 'default'}

        pol.del_rule(rule)

        self.assertEqual(pol._rules, {'other': 'rule2'})
        self.assertEqual(pol._defaults, {'name': 'default'})

    @mock.patch.object(rules, 'RuleDoc', return_value='doc')
    def test_get_doc_exists(self, mock_RuleDoc):
        pol = policy.Policy()
        pol._docs['name'] = 'other'

        result = pol.get_doc('name')

        self.assertEqual(result, 'other')
        self.assertEqual(pol._docs, {'name': 'other'})
        self.assertFalse(mock_RuleDoc.called)

    @mock.patch.object(rules, 'RuleDoc', return_value='doc')
    def test_get_doc_empty(self, mock_RuleDoc):
        pol = policy.Policy()

        result = pol.get_doc('name')

        self.assertEqual(result, 'doc')
        self.assertEqual(pol._docs, {'name': 'doc'})
        mock_RuleDoc.assert_called_once_with('name')

    def test_get_docs(self):
        pol = policy.Policy()
        pol._docs = {'a': 'one', 'b': 'two', 'c': 'three'}

        result = pol.get_docs()

        self.assertEqual(set(result), set(['one', 'two', 'three']))

    @mock.patch.object(pkg_resources, 'iter_entry_points', return_value=[
        mock.Mock(**{'load.side_effect': ImportError()}),
        mock.Mock(**{'load.side_effect': AttributeError()}),
        mock.Mock(**{'load.side_effect': pkg_resources.UnknownExtra()}),
        mock.Mock(**{'load.return_value': 'value1'}),
        mock.Mock(**{'load.return_value': 'value2'}),
    ])
    def test_resolve_builtin(self, mock_iter_entry_points):
        pol = policy.Policy()

        result = pol.resolve('abs')

        self.assertEqual(result, abs)
        self.assertFalse(mock_iter_entry_points.called)

    @mock.patch.object(pkg_resources, 'iter_entry_points', return_value=[
        mock.Mock(**{'load.side_effect': ImportError()}),
        mock.Mock(**{'load.side_effect': AttributeError()}),
        mock.Mock(**{'load.side_effect': pkg_resources.UnknownExtra()}),
        mock.Mock(**{'load.return_value': 'value1'}),
        mock.Mock(**{'load.return_value': 'value2'}),
    ])
    def test_resolve_nogroup(self, mock_iter_entry_points):
        pol = policy.Policy()

        result = pol.resolve('other')

        self.assertEqual(result, None)
        self.assertEqual(pol._resolve_cache['other'], None)
        self.assertFalse(mock_iter_entry_points.called)

    @mock.patch.object(pkg_resources, 'iter_entry_points', return_value=[
        mock.Mock(**{'load.side_effect': ImportError()}),
        mock.Mock(**{'load.side_effect': AttributeError()}),
        mock.Mock(**{'load.side_effect': pkg_resources.UnknownExtra()}),
        mock.Mock(**{'load.return_value': 'value1'}),
        mock.Mock(**{'load.return_value': 'value2'}),
    ])
    def test_resolve_withgroup(self, mock_iter_entry_points):
        pol = policy.Policy('group')

        result = pol.resolve('other')

        self.assertEqual(result, 'value1')
        self.assertEqual(pol._resolve_cache['other'], 'value1')
        mock_iter_entry_points.assert_called_once_with('group', 'other')

    @mock.patch.object(pkg_resources, 'iter_entry_points', return_value=[])
    def test_resolve_withgroup_noresolve(self, mock_iter_entry_points):
        pol = policy.Policy('group')

        result = pol.resolve('other')

        self.assertEqual(result, None)
        self.assertEqual(pol._resolve_cache['other'], None)
        mock_iter_entry_points.assert_called_once_with('group', 'other')

    @mock.patch('logging.getLogger')
    @mock.patch('policies.authorization.Authorization', return_value='authz')
    @mock.patch.object(policy.Policy, 'context_class',
                       return_value=mock.Mock(authz='ctxt_authz'))
    def test_evaluate_norule(self, mock_PolicyContext, mock_Authorization,
                             mock_getLogger):
        pol = policy.Policy()

        result = pol.evaluate('name')

        self.assertEqual(result, 'authz')
        mock_Authorization.assert_called_once_with(False)
        self.assertFalse(mock_PolicyContext.called)
        self.assertFalse(mock_getLogger.called)

    @mock.patch('logging.getLogger')
    @mock.patch('policies.authorization.Authorization', return_value='authz')
    @mock.patch.object(policy.Policy, 'context_class',
                       return_value=mock.Mock(**{
                           'authz': 'ctxt_authz',
                           'push_rule.return_value': mock.MagicMock(),
                       }))
    def test_evaluate_default(self, mock_PolicyContext, mock_Authorization,
                              mock_getLogger):
        ctxt = mock_PolicyContext.return_value
        default = mock.Mock(attrs={'a': 1})
        pol = policy.Policy()
        pol._defaults['name'] = default

        result = pol.evaluate('name')

        self.assertEqual(result, 'ctxt_authz')
        self.assertFalse(mock_Authorization.called)
        mock_PolicyContext.assert_called_once_with(pol, {'a': 1}, {})
        ctxt.push_rule.assert_called_once_with('name')
        ctxt.push_rule.return_value.assert_has_calls([
            mock.call.__enter__(),
            mock.call.__exit__(None, None, None),
        ])
        default.instructions.assert_called_once_with(
            mock_PolicyContext.return_value)

    @mock.patch('logging.getLogger')
    @mock.patch('policies.authorization.Authorization', return_value='authz')
    @mock.patch.object(policy.Policy, 'context_class',
                       return_value=mock.Mock(**{
                           'authz': 'ctxt_authz',
                           'push_rule.return_value': mock.MagicMock(),
                       }))
    def test_evaluate_rule(self, mock_PolicyContext, mock_Authorization,
                           mock_getLogger):
        ctxt = mock_PolicyContext.return_value
        rule = mock.Mock(attrs={'a': 1})
        pol = policy.Policy()
        pol._rules['name'] = rule

        result = pol.evaluate('name')

        self.assertEqual(result, 'ctxt_authz')
        self.assertFalse(mock_Authorization.called)
        mock_PolicyContext.assert_called_once_with(pol, {'a': 1}, {})
        ctxt.push_rule.assert_called_once_with('name')
        ctxt.push_rule.return_value.assert_has_calls([
            mock.call.__enter__(),
            mock.call.__exit__(None, None, None),
        ])
        rule.instructions.assert_called_once_with(
            mock_PolicyContext.return_value)

    @mock.patch('logging.getLogger')
    @mock.patch('policies.authorization.Authorization', return_value='authz')
    @mock.patch.object(policy.Policy, 'context_class',
                       return_value=mock.Mock(**{
                           'authz': 'ctxt_authz',
                           'push_rule.return_value': mock.MagicMock(),
                       }))
    def test_evaluate_both(self, mock_PolicyContext, mock_Authorization,
                           mock_getLogger):
        ctxt = mock_PolicyContext.return_value
        rule = mock.Mock(attrs={'a': 1, 'b': 2})
        default = mock.Mock(attrs={'b': 3, 'c': 4})
        pol = policy.Policy()
        pol._rules['name'] = rule
        pol._defaults['name'] = default

        result = pol.evaluate('name')

        self.assertEqual(result, 'ctxt_authz')
        self.assertFalse(mock_Authorization.called)
        mock_PolicyContext.assert_called_once_with(
            pol, {'a': 1, 'b': 2, 'c': 4}, {})
        ctxt.push_rule.assert_called_once_with('name')
        ctxt.push_rule.return_value.assert_has_calls([
            mock.call.__enter__(),
            mock.call.__exit__(None, None, None),
        ])
        rule.instructions.assert_called_once_with(
            mock_PolicyContext.return_value)
        self.assertFalse(default.instructions.called)

    @mock.patch('logging.getLogger')
    @mock.patch('policies.authorization.Authorization', return_value='authz')
    @mock.patch.object(policy.Policy, 'context_class',
                       return_value=mock.Mock(**{
                           'authz': 'ctxt_authz',
                           'push_rule.return_value': mock.MagicMock(),
                       }))
    def test_evaluate_variables(self, mock_PolicyContext, mock_Authorization,
                                mock_getLogger):
        ctxt = mock_PolicyContext.return_value
        rule = mock.Mock(attrs={'a': 1})
        pol = policy.Policy()
        pol._rules['name'] = rule

        result = pol.evaluate('name', {'x': 23, 'y': 24, 'z': 25})

        self.assertEqual(result, 'ctxt_authz')
        self.assertFalse(mock_Authorization.called)
        mock_PolicyContext.assert_called_once_with(
            pol, {'a': 1}, {'x': 23, 'y': 24, 'z': 25})
        ctxt.push_rule.assert_called_once_with('name')
        ctxt.push_rule.return_value.assert_has_calls([
            mock.call.__enter__(),
            mock.call.__exit__(None, None, None),
        ])
        rule.instructions.assert_called_once_with(
            mock_PolicyContext.return_value)

    @mock.patch('logging.getLogger')
    @mock.patch('policies.authorization.Authorization', return_value='authz')
    @mock.patch.object(policy.Policy, 'context_class',
                       return_value=mock.Mock(**{
                           'authz': 'ctxt_authz',
                           'push_rule.return_value': mock.MagicMock(),
                       }))
    def test_evaluate_exception(self, mock_PolicyContext, mock_Authorization,
                                mock_getLogger):
        ctxt = mock_PolicyContext.return_value
        rule = mock.Mock(**{
            'attrs': {'a': 1},
            'instructions.side_effect': tests.TestException("test"),
        })
        pol = policy.Policy()
        pol._rules['name'] = rule

        result = pol.evaluate('name')

        self.assertEqual(result, 'authz')
        mock_Authorization.assert_called_once_with(False, {'a': 1})
        mock_PolicyContext.assert_called_once_with(pol, {'a': 1}, {})
        ctxt.push_rule.assert_called_once_with('name')
        ctxt.push_rule.return_value.assert_has_calls([
            mock.call.__enter__(),
            mock.call.__exit__(tests.TestException, mock.ANY, mock.ANY),
        ])
        rule.instructions.assert_called_once_with(
            mock_PolicyContext.return_value)


class TestWantContext(tests.TestCase):
    def test_decorator(self):
        def func():
            pass

        result = policy.want_context(func)

        self.assertEqual(result, func)
        self.assertEqual(result._policies_want_context, True)


class TestRule(tests.TestCase):
    @mock.patch('logging.getLogger')
    def test_cached(self, mock_getLogger):
        rule = mock.Mock()
        ctxt = mock.Mock(**{
            'push_rule': mock.MagicMock(),
            'rule_cache': {'name': 'cached'},
            'policy': {'name': rule},
            'stack': [],
        })
        ctxt.name = 'parent'

        policy.rule(ctxt, 'name')

        self.assertFalse(rule.instructions.called)
        self.assertFalse(ctxt.push_rule.called)
        self.assertEqual(ctxt.rule_cache, {'name': 'cached'})
        self.assertEqual(ctxt.stack, ['cached'])
        self.assertFalse(mock_getLogger.called)

    @mock.patch('logging.getLogger')
    def test_missing_rule(self, mock_getLogger):
        ctxt = mock.Mock(**{
            'push_rule': mock.MagicMock(),
            'rule_cache': {},
            'policy': {},
            'stack': [],
        })
        ctxt.name = 'parent'

        policy.rule(ctxt, 'name')

        self.assertFalse(ctxt.push_rule.called)
        self.assertEqual(ctxt.rule_cache, {'name': False})
        self.assertEqual(ctxt.stack, [False])
        mock_getLogger.assert_called_once_with('policies')
        mock_getLogger.return_value.warn.assert_called_once_with(
            "Request to evaluate non-existant rule 'name' while evaluating "
            "rule 'parent'")

    @mock.patch('logging.getLogger')
    def test_success(self, mock_getLogger):
        rule = mock.Mock(**{
            'instructions.side_effect': lambda c, s: c.stack.append('spam'),
        })
        ctxt = mock.Mock(**{
            'push_rule': mock.MagicMock(),
            'rule_cache': {},
            'policy': {'name': rule},
            'stack': [],
        })
        ctxt.name = 'parent'

        policy.rule(ctxt, 'name')

        rule.instructions.assert_called_once_with(ctxt, True)
        ctxt.push_rule.assert_called_once_with('name')
        ctxt.push_rule.return_value.assert_has_calls([
            mock.call.__enter__(),
            mock.call.__exit__(None, None, None),
        ])
        self.assertEqual(ctxt.rule_cache, {'name': 'spam'})
        self.assertEqual(ctxt.stack, ['spam'])
        self.assertFalse(mock_getLogger.called)



# Copyright (C) 2013 by Kevin L. Mitchell <klmitch@mit.edu>
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

import abc
import operator

import six

from policies import authorization


__all__ = ['Instructions',
           'Constant', 'Attribute', 'Ident', 'SetOperator', 'CallOperator',
           'AuthorizationAttr',
           'inv_op', 'pos_op', 'neg_op', 'not_op',
           'pow_op', 'mul_op', 'true_div_op', 'floor_div_op', 'mod_op',
           'add_op', 'sub_op', 'left_shift_op', 'right_shift_op',
           'bit_and_op', 'bit_xor_op', 'bit_or_op', 'in_op', 'not_in_op',
           'is_op', 'is_not_op', 'lt_op', 'gt_op', 'le_op', 'ge_op', 'ne_op',
           'eq_op', 'and_op', 'or_op',
           'item_op',
           'trinary_op',
           'set_authz']


@six.add_metaclass(abc.ABCMeta)
class AbstractInstruction(object):
    """
    An instruction is a callable that manipulates the evaluation
    context.  A sequence of instructions describes how to evaluate an
    expression using a value stack available in the evaluation
    context.
    """

    def __ne__(self, other):
        """
        Compare two instructions for inequivalence.

        :param other: Another ``AbstractInstruction`` to compare to.

        :returns: A ``False`` value if the ``other`` instruction is
                  equivalent to this one, ``True`` otherwise.
        """

        return not self.__eq__(other)

    @abc.abstractmethod
    def __repr__(self):
        """
        Return a representation of this instruction.  Should provide
        enough information for a user to understand what operation
        will be performed.

        :returns: A string representation of this instruction.
        """

        pass  # pragma: nocover

    @abc.abstractmethod
    def __call__(self, ctxt):
        """
        Evaluate this instruction.  Should update the evaluation
        context stack.

        :param ctxt: The evaluation context.
        """

        pass  # pragma: nocover

    @abc.abstractmethod
    def __hash__(self, *elems):
        """
        Return a hash value for this instruction.  Extra parameters
        can be passed in by subclasses to allow those elements to also
        factor into the hash value.

        :returns: The hash value.
        """

        return hash((self.__class__,) + elems)

    @abc.abstractmethod
    def __eq__(self, other):
        """
        Compare two instructions for equivalence.

        :param other: Another ``AbstractInstruction`` to compare to.

        :returns: A ``True`` value if the ``other`` instruction is
                  equivalent to this one, ``False`` otherwise.
        """

        return self.__class__ is other.__class__


class Instructions(AbstractInstruction):
    """
    A special instruction that actually consists of a list of
    instructions.  This is needed for collapsing token lists in the
    parser, and also serves as the return value for the expression
    parser.
    """

    def __init__(self, instructions):
        """
        Initialize an ``Instructions`` object.

        :param instructions: A sequence of the instructions to be
                             contained by the ``Instructions`` object.
        """

        # Linearize the instructions into a flat tuple
        self.instructions = tuple(self._linearize(instructions))

    def __repr__(self):
        """
        Return a representation of this instruction.  Should provide
        enough information for a user to understand what operation
        will be performed.

        :returns: A string representation of this instruction.
        """

        return "Instructions(%r)" % (self.instructions,)

    def __call__(self, ctxt, no_authz=False):
        """
        Evaluate this instruction.  Executes the contained
        instructions in sequence.

        :param ctxt: The evaluation context.
        :param no_authz: If ``True``, evaluation will stop at the
                         set_authz instruction.  This can be used to
                         only evaluate the expression in a rule.
        """

        for inst in self.instructions:
            # Allows for evaluating only the expression of a rule
            if no_authz and inst == set_authz:
                break

            inst(ctxt)

    def __hash__(self):
        """
        Return a hash value for this instruction.

        :returns: The hash value.
        """

        return super(Instructions, self).__hash__(*self.instructions)

    def __eq__(self, other):
        """
        Compare two instructions for equivalence.

        :param other: Another ``AbstractInstruction`` to compare to.

        :returns: A ``True`` value if the ``other`` instruction is
                  equivalent to this one, ``False`` otherwise.
        """

        return (super(Instructions, self).__eq__(other) and
                self.instructions == other.instructions)

    @classmethod
    def _linearize(cls, inst_list):
        """
        A generator function which performs linearization of the list
        of instructions; that is, each instruction which should be
        executed will be yielded in turn, recursing into
        ``Instructions`` instances that appear in the list.

        :param inst_list: A list (or other sequence) of instructions.

        :returns: An iterator which returns all instructions.
        """

        for inst in inst_list:
            # Check if we need to recurse
            if isinstance(inst, Instructions):
                for sub_inst in cls._linearize(inst.instructions):
                    yield sub_inst
            else:
                yield inst


class Constant(AbstractInstruction):
    """
    An instruction that pushes a constant value onto the evaluation
    context stack.
    """

    def __init__(self, value):
        """
        Initialize a ``Constant`` object.

        :param value: The value of the constant.
        """

        self.value = value

    def __repr__(self):
        """
        Return a representation of this instruction.  Should provide
        enough information for a user to understand what operation
        will be performed.

        :returns: A string representation of this instruction.
        """

        return "Constant(%r)" % (self.value,)

    def __call__(self, ctxt):
        """
        Evaluate this instruction.  Pushes the value of the constant
        onto the evaluation context stack.

        :param ctxt: The evaluation context.
        """

        ctxt.stack.append(self.value)

    def __hash__(self):
        """
        Return a hash value for this instruction.

        :returns: The hash value.
        """

        return super(Constant, self).__hash__(self.value)

    def __eq__(self, other):
        """
        Compare two instructions for equivalence.

        :param other: Another ``AbstractInstruction`` to compare to.

        :returns: A ``True`` value if the ``other`` instruction is
                  equivalent to this one, ``False`` otherwise.
        """

        return (super(Constant, self).__eq__(other) and
                self.value == other.value)


class Attribute(AbstractInstruction):
    """
    An instruction that replaces the top of the evaluation context
    stack with one of its attributes.
    """

    def __init__(self, attribute):
        """
        Initialize an ``Attribute`` object.

        :param attribute: The name of the attribute.
        """

        self.attribute = attribute

    def __repr__(self):
        """
        Return a representation of this instruction.  Should provide
        enough information for a user to understand what operation
        will be performed.

        :returns: A string representation of this instruction.
        """

        return "Attribute(%r)" % (self.attribute,)

    def __call__(self, ctxt):
        """
        Evaluate this instruction.  Pops a value off the top of the
        evaluation context stack and pushes the value of its
        attribute.

        :param ctxt: The evaluation context.
        """

        ctxt.stack.append(getattr(ctxt.stack.pop(), self.attribute))

    def __hash__(self):
        """
        Return a hash value for this instruction.

        :returns: The hash value.
        """

        return super(Attribute, self).__hash__(self.attribute)

    def __eq__(self, other):
        """
        Compare two instructions for equivalence.

        :param other: Another ``AbstractInstruction`` to compare to.

        :returns: A ``True`` value if the ``other`` instruction is
                  equivalent to this one, ``False`` otherwise.
        """

        return (super(Attribute, self).__eq__(other) and
                self.attribute == other.attribute)


class Ident(AbstractInstruction):
    """
    An instruction that resolves an identifier into a value and pushes
    that value onto the evaluation context stack.
    """

    def __init__(self, ident):
        """
        Initialize an ``Ident`` object.

        :param ident: The identifier.
        """

        self.ident = ident

    def __repr__(self):
        """
        Return a representation of this instruction.  Should provide
        enough information for a user to understand what operation
        will be performed.

        :returns: A string representation of this instruction.
        """

        return "Ident(%r)" % (self.ident,)

    def __call__(self, ctxt):
        """
        Evaluate this instruction.  Resolves the identifier in the
        evaluation context and pushes its value onto the evaluation
        context stack.

        :param ctxt: The evaluation context.
        """

        ctxt.stack.append(ctxt.resolve(self.ident))

    def __hash__(self):
        """
        Return a hash value for this instruction.

        :returns: The hash value.
        """

        return super(Ident, self).__hash__(self.ident)

    def __eq__(self, other):
        """
        Compare two instructions for equivalence.

        :param other: Another ``AbstractInstruction`` to compare to.

        :returns: A ``True`` value if the ``other`` instruction is
                  equivalent to this one, ``False`` otherwise.
        """

        return (super(Ident, self).__eq__(other) and
                self.ident == other.ident)


class Operator(AbstractInstruction):
    """
    An instruction that performs an operation on some elements of the
    evaluation context stack, replacing those elements with the return
    value of the operation.
    """

    def __init__(self, count, opstr):
        """
        Initialize an ``Operator`` object.

        :param count: The number of stack elements that will be
                      consumed by the operator.
        :param opstr: A string representing the operation that will be
                      performed.  This is used when a representation
                      of this operator is requested.
        """

        self.count = count
        self.opstr = opstr

    def __repr__(self):
        """
        Return a representation of this instruction.  Should provide
        enough information for a user to understand what operation
        will be performed.

        :returns: A string representation of this instruction.
        """

        return '%s(%d, %r)' % (self.__class__.__name__, self.count, self.opstr)

    def __call__(self, ctxt):
        """
        Evaluate this instruction.  Replaces the ``count`` elements on
        the top of the evaluation context stack with the single
        element obtained by calling ``op()`` on those elements.

        :param ctxt: The evaluation context.
        """

        ctxt.stack[-self.count:] = [self.op(*ctxt.stack[-self.count:])]

    def __hash__(self, *elems):
        """
        Return a hash value for this instruction.  Extra parameters
        can be passed in by subclasses to allow those elements to also
        factor into the hash value.

        :returns: The hash value.
        """

        return super(Operator, self).__hash__(self.count, *elems)

    def __eq__(self, other):
        """
        Compare two instructions for equivalence.

        :param other: Another ``AbstractInstruction`` to compare to.

        :returns: A ``True`` value if the ``other`` instruction is
                  equivalent to this one, ``False`` otherwise.
        """

        return (super(Operator, self).__eq__(other) and
                self.count == other.count)

    def fold(self, elems):
        """
        Perform constant folding.  If all the elements are constant,
        then the operator is performed and the constant is folded.

        :param elems: A list (or list-like object) containing the
                      elements.

        :returns: If all elements are instances of ``Constant``,
                  returns a list of one element, which is the result
                  of calling the operation on ``elems``.  Otherwise,
                  returns a list composed of the elements of ``elems``
                  plus this operator.
        """

        # Are the elements constants?
        if all(isinstance(e, Constant) for e in elems):
            return [Constant(self.op(*[e.value for e in elems]))]

        return [Instructions(elems[:] + [self])]

    @abc.abstractmethod
    def op(self, *args):
        """
        Perform the operation.  Will be passed the stack elements.

        :returns: The desired replacement value.
        """

        pass  # pragma: nocover


class GenericOperator(Operator):
    """
    A generic operator which evaluates a callable passed to the
    constructor.
    """

    def __init__(self, count, op, opstr):
        """
        Initialize a ``GenericOperator`` object.

        :param count: The number of stack elements that will be
                      consumed by the operator.
        :param op: The callable implementing the operation.
        :param opstr: A string representing the operation that will be
                      performed.  This is used when a representation
                      of this operator is requested.
        """

        super(GenericOperator, self).__init__(count, opstr)
        self._op = op

    def __hash__(self):
        """
        Return a hash value for this instruction.

        :returns: The hash value.
        """

        return super(GenericOperator, self).__hash__(self._op)

    def __eq__(self, other):
        """
        Compare two instructions for equivalence.

        :param other: Another ``AbstractInstruction`` to compare to.

        :returns: A ``True`` value if the ``other`` instruction is
                  equivalent to this one, ``False`` otherwise.
        """

        return (super(GenericOperator, self).__eq__(other) and
                self._op == other._op)

    def op(self, *args):
        """
        Call the operation that was passed to the constructor with the
        given arguments.

        :returns: The result of calling the operator.
        """

        return self._op(*args)


class TrinaryOperator(Operator):
    """
    An instruction that implements the trinary operation.  If the
    conditional evaluates to ``True``, the ``if_true`` element is
    returned, otherwise the ``if_false`` element is returned.  The
    three elements are drawn from the stack in the indicated order.
    """

    def __init__(self):
        """
        Initialize a ``TrinaryOperator`` object.
        """

        super(TrinaryOperator, self).__init__(3, 'if/else')

    def op(self, cond, if_true, if_false):
        """
        Implement the trinary operation.

        :param cond: The conditional.
        :param if_true: The value to return if the conditional
                        evaluates to ``True``.
        :param if_false: The value to return if the conditional
                         evaluates to ``False``.

        :returns: One of ``if_true`` or ``if_false``, depending on the
                  boolean value of ``cond``.
        """

        return if_true if cond else if_false

    def fold(self, elems):
        """
        Perform constant folding.  Unlike ``Operator.fold()``, the
        only important operand for determining whether the expression
        can be folded is the value of the conditional.  If it is a
        constant value, then the trinary expression can be evaluated
        immediately.

        :param elems: A list (or list-like object) containing the
                      elements.

        :returns: If the first element is an instance of ``Constant``,
                  returns either the second or third element,
                  depending on the boolean value of the first element.
                  Otherwise, returns a list composed of the elements
                  of ``elems`` plus this operator.
        """

        if isinstance(elems[0], Constant):
            return [elems[1] if elems[0].value else elems[2]]

        return [Instructions(elems[:] + [self])]


class SetOperator(Operator):
    """
    An instruction that constructs a frozen set from a given number of
    elements on the evaluation context stack, replacing those elements
    with the set.
    """

    def __init__(self, count):
        """
        Initialize a ``SetOperator`` object.

        :param count: The number of elements to construct the set
                      from.
        """

        super(SetOperator, self).__init__(count, 'set')

    def op(self, *args):
        """
        Construct a set with the given arguments.

        :returns: The constructed set.
        """

        return frozenset(args)


class CallOperator(AbstractInstruction):
    """
    An instruction that performs a function or method call.  The top
    ``count`` elements on the stack identify the function or method
    and its arguments.  The function or method reference and the
    arguments will be replaced by the return value of calling the
    function or method.
    """

    def __init__(self, count):
        """
        Initialize a ``CallOperator`` object.

        :param count: The number of elements to construct the call
                      from.  The first element is the callable, and
                      the remaining elements will be positional
                      arguments to the callable.
        """

        self.count = count

    def __repr__(self):
        """
        Return a representation of this instruction.  Should provide
        enough information for a user to understand what operation
        will be performed.

        :returns: A string representation of this instruction.
        """

        return 'CallOperator(%d)' % self.count

    def __call__(self, ctxt):
        """
        Evaluate this instruction.  Replaces the ``count`` elements on
        the top of the evaluation context stack with the single
        element obtained by calling the function on its arguments.

        :param ctxt: The evaluation context.
        """

        # Get the function and its arguments
        args = ctxt.stack[-self.count:]
        func = args.pop(0)

        # If the function wants the context, add the context and call
        # it; it is assumed the function will do its own updates to
        # the context stack
        if getattr(func, '_policies_want_context', False):
            ctxt.stack = ctxt.stack[:-self.count]
            func(ctxt, *args)
        else:
            # Call the function and update the stack
            ctxt.stack[-self.count:] = [func(*args)]

    def __hash__(self):
        """
        Return a hash value for this instruction.

        :returns: The hash value.
        """

        return super(CallOperator, self).__hash__(self.count)

    def __eq__(self, other):
        """
        Compare two instructions for equivalence.

        :param other: Another ``AbstractInstruction`` to compare to.

        :returns: A ``True`` value if the ``other`` instruction is
                  equivalent to this one, ``False`` otherwise.
        """

        return (super(CallOperator, self).__eq__(other) and
                self.count == other.count)


class SetAuthorization(AbstractInstruction):
    """
    An instruction that sets the result of the authorization check.
    This initializes the ``authz`` attribute of the evaluation context
    based on the boolean value of the top of the stack.
    """

    def __repr__(self):
        """
        Return a representation of this instruction.  Should provide
        enough information for a user to understand what operation
        will be performed.

        :returns: A string representation of this instruction.
        """

        return 'SetAuthorization()'

    def __call__(self, ctxt):
        """
        Evaluate this instruction.  Creates the ``authz`` attribute of
        the evaluation context, with the boolean value of the top of
        the stack.  The attribute defaults are drawn from the
        ``attrs`` attribute of the evaluation context, which is
        expected to be a dictionary.

        :param ctxt: The evaluation context.
        """

        ctxt.authz = authorization.Authorization(ctxt.stack.pop(), ctxt.attrs)

    def __hash__(self):
        """
        Return a hash value for this instruction.

        :returns: The hash value.
        """

        return super(SetAuthorization, self).__hash__()

    def __eq__(self, other):
        """
        Compare two instructions for equivalence.

        :param other: Another ``AbstractInstruction`` to compare to.

        :returns: A ``True`` value if the ``other`` instruction is
                  equivalent to this one, ``False`` otherwise.
        """

        return super(SetAuthorization, self).__eq__(other)


class AuthorizationAttr(AbstractInstruction):
    """
    An instruction that sets attributes on the authorization result.
    The ``SetAuthorization`` instruction MUST be executed before this
    instruction can be called.  A value will be popped off the stack
    and its value set on a named attribute of the authorization
    result.
    """

    def __init__(self, attribute):
        """
        Initialize an ``AuthorizationAttr`` object.

        :param attribute: The name of the attribute to set.
        """

        self.attribute = attribute

    def __repr__(self):
        """
        Return a representation of this instruction.  Should provide
        enough information for a user to understand what operation
        will be performed.

        :returns: A string representation of this instruction.
        """

        return 'AuthorizationAttr(%r)' % self.attribute

    def __call__(self, ctxt):
        """
        Evaluate this instruction.  Pops a value off the top of the
        evaluation context stack and sets the corresponding attribute
        of the authorization result to that value.

        :param ctxt: The evaluation context.
        """

        ctxt.authz._attrs[self.attribute] = ctxt.stack.pop()

    def __hash__(self):
        """
        Return a hash value for this instruction.

        :returns: The hash value.
        """

        return super(AuthorizationAttr, self).__hash__(self.attribute)

    def __eq__(self, other):
        """
        Compare two instructions for equivalence.

        :param other: Another ``AbstractInstruction`` to compare to.

        :returns: A ``True`` value if the ``other`` instruction is
                  equivalent to this one, ``False`` otherwise.
        """

        return (super(AuthorizationAttr, self).__eq__(other) and
                self.attribute == other.attribute)


# Unary operators
inv_op = GenericOperator(1, operator.inv, '~')
pos_op = GenericOperator(1, operator.pos, '+')
neg_op = GenericOperator(1, operator.neg, '-')
not_op = GenericOperator(1, operator.not_, 'not')

# Binary operators
pow_op = GenericOperator(2, operator.pow, '**')
mul_op = GenericOperator(2, operator.mul, '*')
true_div_op = GenericOperator(2, operator.truediv, '/')
floor_div_op = GenericOperator(2, operator.floordiv, '//')
mod_op = GenericOperator(2, operator.mod, '%')
add_op = GenericOperator(2, operator.add, '+')
sub_op = GenericOperator(2, operator.sub, '-')
left_shift_op = GenericOperator(2, operator.lshift, '<<')
right_shift_op = GenericOperator(2, operator.rshift, '>>')
bit_and_op = GenericOperator(2, operator.and_, '&')
bit_xor_op = GenericOperator(2, operator.xor, '^')
bit_or_op = GenericOperator(2, operator.or_, '|')
in_op = GenericOperator(2, lambda x, y: x in y, 'in')
not_in_op = GenericOperator(2, lambda x, y: x not in y, 'not in')
is_op = GenericOperator(2, operator.is_, 'is')
is_not_op = GenericOperator(2, operator.is_not, 'is not')
lt_op = GenericOperator(2, operator.lt, '<')
gt_op = GenericOperator(2, operator.gt, '>')
le_op = GenericOperator(2, operator.le, '<=')
ge_op = GenericOperator(2, operator.ge, '>=')
ne_op = GenericOperator(2, operator.ne, '!=')
eq_op = GenericOperator(2, operator.eq, '==')
and_op = GenericOperator(2, lambda x, y: x and y, 'and')
or_op = GenericOperator(2, lambda x, y: x or y, 'or')
item_op = GenericOperator(2, lambda x, y: x[y], '[]')

# The trinary operator
trinary_op = TrinaryOperator()

# The set authorization instruction
set_authz = SetAuthorization()



#!/usr/bin/env python

import os

from setuptools import setup


def readreq(filename):
    result = []
    with open(filename) as f:
        for req in f:
            req = req.lstrip()
            if req.startswith('-e ') or req.startswith('http:'):
                idx = req.find('#egg=')
                if idx >= 0:
                    req = req[idx + 5:].partition('#')[0].strip()
                else:
                    pass
            else:
                req = req.partition('#')[0].strip()
            if not req:
                continue
            result.append(req)
    return result


def readfile(filename):
    with open(filename) as f:
        return f.read()


setup(
    name='policies',
    version='0.3',
    author='Kevin L. Mitchell',
    author_email='klmitch@mit.edu',
    url='https://github.com/klmitch/policies',
    description="An access policy language evaluator.",
    long_description=readfile('README.rst'),
    license='GNU General Public License v3 or later (GPLv3+)',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 '
            'or later (GPLv3+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Security',
        'Topic :: Software Development :: Interpreters',
        'Topic :: Software Development :: Libraries',
    ],
    packages=['policies'],
    install_requires=readreq('requirements.txt'),
    tests_require=readreq('test-requirements.txt'),
)



# Copyright (C) 2013 by Kevin L. Mitchell <klmitch@mit.edu>
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

import logging

import pyparsing
import six

from policies.instructions import *


pyparsing.ParserElement.enablePackrat()


def unary_construct(tokens):
    """
    Construct proper instructions for unary expressions.  For
    instance, if the tokens represent "~ 1", this will return the
    instruction array "1 inv_op".

    :param tokens: The sequence of tokens.

    :returns: An instance of ``Instructions`` containing the list of
              instructions.
    """

    op, operand = tokens

    return op.fold([operand])


def binary_construct(tokens):
    """
    Construct proper instructions for binary expressions from a
    sequence of tokens at the same precedence level.  For instance, if
    the tokens represent "1 + 2 + 3", this will return the instruction
    array "1 2 add_op 3 add_op".

    :param tokens: The sequence of tokens.

    :returns: An instance of ``Instructions`` containing the list of
              instructions.
    """

    # Initialize the list of instructions we will return with the
    # left-most element
    instructions = [tokens[0]]

    # Now process all the remaining tokens, building up the array we
    # will return
    for i in range(1, len(tokens), 2):
        op, rhs = tokens[i:i + 2]

        # Add the right-hand side
        instructions.append(rhs)

        # Now apply constant folding
        instructions[-2:] = op.fold(instructions[-2:])

    return instructions


def str_decode(string):
    """
    Decode escapes in a string.

    :param string: The string to decode the escapes in.

    :returns: The decoded string.
    """

    if six.PY2:  # pragma: nocover
        return string.decode('unicode-escape')
    else:  # pragma: nocover
        return bytes(string, 'utf-8').decode('unicode-escape')


# Primitive values
TRUE = pyparsing.Keyword('True').setParseAction(lambda: [Constant(True)])
FALSE = pyparsing.Keyword('False').setParseAction(lambda: [Constant(False)])
NONE = pyparsing.Keyword('None').setParseAction(lambda: [Constant(None)])
INT = (
    pyparsing.Regex(r'[+-]?\d+') +
    ~pyparsing.FollowedBy(pyparsing.Regex(r'[.eE]'))
).setParseAction(lambda t: [Constant(int(t[0]))])
FLOAT = (
    pyparsing.Regex(r'[+-]?\d+(?:\.\d*)?(?:[eE][+-]?\d+)?')
).setParseAction(lambda t: [Constant(float(t[0]))])
STR = (
    pyparsing.OneOrMore(pyparsing.quotedString)
).setParseAction(lambda t: [Constant(''.join(str_decode(v[1:-1]) for v in t))])
IDENT = (
    pyparsing.Regex(r'[a-zA-Z_][a-zA-Z0-9_]*')
).setParseAction(lambda t: [Ident(t[0])])
ATTR = (
    pyparsing.Regex(r'[a-zA-Z_][a-zA-Z0-9_]*')
).setParseAction(lambda t: [Attribute(t[0])])
AUTHATTR = (
    pyparsing.Regex(r'[a-zA-Z][a-zA-Z0-9_]*')
).setParseAction(lambda t: [AuthorizationAttr(t[0])])

# Unary operators
INV_OP = pyparsing.Literal('~').setParseAction(lambda: [inv_op])
POS_OP = pyparsing.Literal('+').setParseAction(lambda: [pos_op])
NEG_OP = pyparsing.Literal('-').setParseAction(lambda: [neg_op])
NOT_OP = pyparsing.Keyword('not').setParseAction(lambda: [not_op])

# Binary operators
POW_OP = pyparsing.Literal('**').setParseAction(lambda: [pow_op])
MUL_OP = pyparsing.Literal('*').setParseAction(lambda: [mul_op])
TRUE_DIV_OP = pyparsing.Literal('/').setParseAction(lambda: [true_div_op])
FLOOR_DIV_OP = pyparsing.Literal('//').setParseAction(lambda: [floor_div_op])
MOD_OP = pyparsing.Literal('%').setParseAction(lambda: [mod_op])
ADD_OP = pyparsing.Literal('+').setParseAction(lambda: [add_op])
SUB_OP = pyparsing.Literal('-').setParseAction(lambda: [sub_op])
LEFT_SHIFT_OP = pyparsing.Literal('<<').setParseAction(lambda: [left_shift_op])
RIGHT_SHIFT_OP = (
    pyparsing.Literal('>>')
).setParseAction(lambda: [right_shift_op])
BIT_AND_OP = pyparsing.Literal('&').setParseAction(lambda: [bit_and_op])
BIT_XOR_OP = pyparsing.Literal('^').setParseAction(lambda: [bit_xor_op])
BIT_OR_OP = pyparsing.Literal('|').setParseAction(lambda: [bit_or_op])
IN_OP = pyparsing.Keyword('in').setParseAction(lambda: [in_op])
NOT_IN_OP = (
    pyparsing.Keyword('not') + pyparsing.Keyword('in')
).setParseAction(lambda: [not_in_op])
IS_OP = pyparsing.Keyword('is').setParseAction(lambda: [is_op])
IS_NOT_OP = (
    pyparsing.Keyword('is') + pyparsing.Keyword('not')
).setParseAction(lambda: [is_not_op])
LT_OP = pyparsing.Literal('<').setParseAction(lambda: [lt_op])
GT_OP = pyparsing.Literal('>').setParseAction(lambda: [gt_op])
LE_OP = pyparsing.Literal('<=').setParseAction(lambda: [le_op])
GE_OP = pyparsing.Literal('>=').setParseAction(lambda: [ge_op])
NE_OP = pyparsing.Literal('!=').setParseAction(lambda: [ne_op])
EQ_OP = pyparsing.Literal('==').setParseAction(lambda: [eq_op])
AND_OP = pyparsing.Keyword('and').setParseAction(lambda: [and_op])
OR_OP = pyparsing.Keyword('or').setParseAction(lambda: [or_op])

# Trinary operators
IF_OP = pyparsing.Keyword('if').suppress()
ELSE_OP = pyparsing.Keyword('else').suppress()

# Useful tokens
LPAREN = pyparsing.Suppress('(')
RPAREN = pyparsing.Suppress(')')
LBRACKET = pyparsing.Suppress('[')
RBRACKET = pyparsing.Suppress(']')
LBRACE = pyparsing.Suppress('{')
RBRACE = pyparsing.Suppress('}')
LLBRACE = pyparsing.Suppress('{{')
RRBRACE = pyparsing.Suppress('}}')
COMMA = pyparsing.Suppress(',')
DOT = pyparsing.Suppress('.')
ASSIGN = pyparsing.Suppress('=')


# Hand-construct the expression.  We have to do this instead of using
# pyparsing.infixNotation(), because we need some special expressions
# at the top precedence level: attribute access, item access, and
# function call.  This internal expression does not have the parse
# action set that consolidates the result as an instance of
# Instructions, because this can mess up constant
# folding--particularly for set literals.
int_expr = pyparsing.Forward()

# Describe set literals
set_literal = (
    LBRACE +
    pyparsing.Optional(int_expr + pyparsing.ZeroOrMore(COMMA + int_expr)) +
    pyparsing.Optional(COMMA) +
    RBRACE
).setParseAction(lambda t: SetOperator(len(t)).fold(t))

# Build the value non-terminal
value = (TRUE | FALSE | NONE | INT | FLOAT | STR | IDENT | set_literal)

# Build the primary non-terminal
primary = value | (LPAREN + int_expr + RPAREN)

# Build the sub-elements of the expr0 precedence level: attribute
# access, item access, and function call
expr0 = pyparsing.Forward()
attr_pattern = DOT + ATTR
item_pattern = (
    LBRACKET + int_expr + RBRACKET
).setParseAction(lambda t: list(t) + [item_op])
call_pattern = (
    LPAREN +
    pyparsing.Optional(int_expr + pyparsing.ZeroOrMore(COMMA + int_expr)) +
    pyparsing.Optional(COMMA) +
    RPAREN
).setParseAction(lambda t: list(t) + [CallOperator(len(t) + 1)])
expr0_pattern = attr_pattern | item_pattern | call_pattern
expr0_match = (
    pyparsing.FollowedBy(primary + expr0_pattern) +
    primary + pyparsing.OneOrMore(expr0_pattern)
).setParseAction(lambda t: [Instructions(t)])
expr0 <<= (expr0_match | primary)

# Build the expr1 precedence level: **
expr1 = pyparsing.Forward()
expr1_ops = POW_OP
expr1_match = (
    pyparsing.FollowedBy(expr0 + expr1_ops + expr1) +
    expr0 + pyparsing.OneOrMore(expr1_ops + expr1)
).setParseAction(binary_construct)
expr1 <<= (expr1_match | expr0)

# Build the expr2 precedence level: unary +, -, and ~
expr2 = pyparsing.Forward()
expr2_ops = INV_OP | POS_OP | NEG_OP
expr2_match = (
    pyparsing.FollowedBy(expr2_ops + expr2) +
    pyparsing.Optional(expr2_ops) + expr2
).setParseAction(unary_construct)
expr2 <<= (expr2_match | expr1)

# Build the expr3 precedence level: arithmetic *, /, //, and %
expr3 = pyparsing.Forward()
# Note: Order here is important; // tried first, then /
expr3_ops = FLOOR_DIV_OP | TRUE_DIV_OP | MUL_OP | MOD_OP
expr3_match = (
    pyparsing.FollowedBy(expr2 + expr3_ops + expr2) +
    expr2 + pyparsing.OneOrMore(expr3_ops + expr2)
).setParseAction(binary_construct)
expr3 <<= (expr3_match | expr2)

# Build the expr4 precedence level: arithmetic + and -
expr4 = pyparsing.Forward()
expr4_ops = ADD_OP | SUB_OP
expr4_match = (
    pyparsing.FollowedBy(expr3 + expr4_ops + expr3) +
    expr3 + pyparsing.OneOrMore(expr4_ops + expr3)
).setParseAction(binary_construct)
expr4 <<= (expr4_match | expr3)

# Build the expr5 precedence level: >> and <<
expr5 = pyparsing.Forward()
expr5_ops = LEFT_SHIFT_OP | RIGHT_SHIFT_OP
expr5_match = (
    pyparsing.FollowedBy(expr4 + expr5_ops + expr4) +
    expr4 + pyparsing.OneOrMore(expr5_ops + expr4)
).setParseAction(binary_construct)
expr5 <<= (expr5_match | expr4)

# Build the expr6 precedence level: &
expr6 = pyparsing.Forward()
expr6_ops = BIT_AND_OP
expr6_match = (
    pyparsing.FollowedBy(expr5 + expr6_ops + expr5) +
    expr5 + pyparsing.OneOrMore(expr6_ops + expr5)
).setParseAction(binary_construct)
expr6 <<= (expr6_match | expr5)

# Build the expr7 precedence level: ^
expr7 = pyparsing.Forward()
expr7_ops = BIT_XOR_OP
expr7_match = (
    pyparsing.FollowedBy(expr6 + expr7_ops + expr6) +
    expr6 + pyparsing.OneOrMore(expr7_ops + expr6)
).setParseAction(binary_construct)
expr7 <<= (expr7_match | expr6)

# Build the expr8 precedence level: |
expr8 = pyparsing.Forward()
expr8_ops = BIT_OR_OP
expr8_match = (
    pyparsing.FollowedBy(expr7 + expr8_ops + expr7) +
    expr7 + pyparsing.OneOrMore(expr8_ops + expr7)
).setParseAction(binary_construct)
expr8 <<= (expr8_match | expr7)

# Build the expr9 precedence level: in, not in, is, is not, <, <=, >,
# >=, !=, and ==
expr9 = pyparsing.Forward()
# Note: Order here is important; e.g., <= tried first, then <
expr9_ops = (IN_OP | NOT_IN_OP | IS_NOT_OP | IS_OP |
             LE_OP | GE_OP | LT_OP | GT_OP | NE_OP | EQ_OP)
expr9_match = (
    pyparsing.FollowedBy(expr8 + expr9_ops + expr8) +
    expr8 + pyparsing.OneOrMore(expr9_ops + expr8)
).setParseAction(binary_construct)
expr9 <<= (expr9_match | expr8)

# Build the expr10 precedence level: unary not
expr10 = pyparsing.Forward()
expr10_ops = NOT_OP
expr10_match = (
    pyparsing.FollowedBy(expr10_ops + expr10) +
    pyparsing.Optional(expr10_ops) + expr10
).setParseAction(unary_construct)
expr10 <<= (expr10_match | expr9)

# Build the expr11 precedence level: logical and
expr11 = pyparsing.Forward()
expr11_ops = AND_OP
expr11_match = (
    pyparsing.FollowedBy(expr10 + expr11_ops + expr10) +
    expr10 + pyparsing.OneOrMore(expr11_ops + expr10)
).setParseAction(binary_construct)
expr11 <<= (expr11_match | expr10)

# Build the expr12 precedence level: logical or
expr12 = pyparsing.Forward()
expr12_ops = OR_OP
expr12_match = (
    pyparsing.FollowedBy(expr11 + expr12_ops + expr11) +
    expr11 + pyparsing.OneOrMore(expr12_ops + expr11)
).setParseAction(binary_construct)
expr12 <<= (expr12_match | expr11)

# Build the expr13 precedence level: trinary if/else
expr13 = pyparsing.Forward()
expr13_ops1 = IF_OP
expr13_ops2 = ELSE_OP
expr13_match = (
    pyparsing.FollowedBy(expr12 + expr13_ops1 + expr12 +
                         expr13_ops2 + expr12) +
    expr12 + expr13_ops1 + expr12 + expr13_ops2 + expr12
).setParseAction(lambda t: trinary_op.fold([t[1], t[0], t[2]]))
expr13 <<= (expr13_match | expr12)

# Finish building internal expressions
int_expr <<= expr13

# Build expressions as used by rules; this adds a parse action that
# wraps up all the instructions into a single instance of Instructions
expr = int_expr('expr')
expr.setParseAction(lambda t: [Instructions(t)])

# Set up authorization attribute assignment; this allows us to set
# arbitrary values on the authorization object, for the benefit of
# client code.  This could be used to restrict access to certain
# fields of a user object, for instance.
assignment = (
    AUTHATTR('attr') + ASSIGN +
    pyparsing.Optional(expr, default=Constant(None))
).setParseAction(lambda t: [Instructions([t['expr'], t['attr']])])

# Finally, we can build a rule
rule = (
    pyparsing.Optional(expr, default=Constant(False)) +
    pyparsing.Optional(
        LLBRACE +
        pyparsing.Optional(
            assignment +
            pyparsing.ZeroOrMore(COMMA + assignment)
        ) +
        pyparsing.Optional(COMMA) +
        RRBRACE
    )
).setParseAction(lambda t: [Instructions([t[0], set_authz] + t[1:])])


def parse_rule(name, rule_text, do_raise=False):
    """
    Parses the given rule text.

    :param name: The name of the rule.  Used when emitting log
                 messages regarding a failure to parse the rule.
    :param rule_text: The text of the rule to parse.
    :param do_raise: If ``False`` and the rule fails to parse, a log
                     message is emitted to the "policies" logger at
                     level WARN, and a rule that always evaluates to
                     ``False`` will be returned.  If ``True``, a
                     ``pyparsing.ParseException`` will be raised.

    :returns: An instance of ``policies.instructions.Instructions``,
              containing the instructions necessary to evaluate the
              authorization rule.
    """

    try:
        return rule.parseString(rule_text, parseAll=True)[0]
    except pyparsing.ParseException as exc:
        # Allow for debugging
        if do_raise:
            raise

        # Get the logger and emit our log messages
        log = logging.getLogger('policies')
        log.warn("Failed to parse rule %r: %s" % (name, exc))
        log.warn("Rule line: %s" % exc.line)
        log.warn("Location : %s^" % (" " * (exc.col - 1)))

        # Construct and return a fail-closed instruction
        return Instructions([Constant(False), set_authz])
